"""
Mock devices that mimic actual data from Abode servers.

This file should be updated any time the Abode server responses
change to confirm this library can still communicate.
"""

EMPTY_DEVICE_RESPONSE = '[]'


def status_put_response_ok(devid, status):
    """Return status change response json."""
    return '{"id": "' + devid + '", "status": "' + str(status) + '"}'


def level_put_response_ok(devid, level):
    """Return level change response json."""
    return '{"id": "' + devid + '", "level": "' + str(level) + '"}'



"""Abode camera device."""
import json
import logging
from shutil import copyfileobj
import requests

from ..exceptions import AbodeException
from ..devices import AbodeDevice
from ..helpers import constants as CONST
from ..helpers import errors as ERROR
from ..helpers import timeline as TIMELINE

_LOGGER = logging.getLogger(__name__)


class AbodeCamera(AbodeDevice):
    """Class to represent a camera device."""

    def __init__(self, json_obj, abode):
        """Set up Abode alarm device."""
        AbodeDevice.__init__(self, json_obj, abode)
        self._image_url = None

    def capture(self):
        """Request a new camera image."""
        # Abode IP cameras use a different URL for image captures.
        if 'control_url_snapshot' in self._json_state:
            url = CONST.BASE_URL + self._json_state['control_url_snapshot']

        elif 'control_url' in self._json_state:
            url = CONST.BASE_URL + self._json_state['control_url']

        else:
            raise AbodeException((ERROR.MISSING_CONTROL_URL))

        try:
            response = self._abode.send_request("put", url)

            _LOGGER.debug("Capture image response: %s", response.text)

            return True

        except AbodeException as exc:
            _LOGGER.warning("Failed to capture image: %s", exc)

        return False

    def refresh_image(self):
        """Get the most recent camera image."""
        url = str.replace(CONST.TIMELINE_IMAGES_ID_URL, '$DEVID$', self.device_id)
        response = self._abode.send_request("get", url)

        _LOGGER.debug("Get image response: %s", response.text)

        return self.update_image_location(json.loads(response.text))

    def update_image_location(self, timeline_json):
        """Update the image location."""
        if not timeline_json:
            return False

        # If we get a list of objects back (likely)
        # then we just want the first one as it should be the "newest"
        if isinstance(timeline_json, (tuple, list)):
            timeline_json = timeline_json[0]

        # Verify that the event code is of the "CAPTURE IMAGE" event
        event_code = timeline_json.get('event_code')
        if event_code != TIMELINE.CAPTURE_IMAGE['event_code']:
            raise AbodeException((ERROR.CAM_TIMELINE_EVENT_INVALID))

        # The timeline response has an entry for "file_path" that acts as the
        # location of the image within the Abode servers.
        file_path = timeline_json.get('file_path')
        if not file_path:
            raise AbodeException((ERROR.CAM_IMAGE_REFRESH_NO_FILE))

        # Perform a "head" request for the image and look for a
        # 302 Found response
        url = CONST.BASE_URL + file_path
        response = self._abode.send_request("head", url)

        if response.status_code != 302:
            _LOGGER.warning(
                "Unexected response code %s with body: %s",
                str(response.status_code),
                response.text,
            )
            raise AbodeException((ERROR.CAM_IMAGE_UNEXPECTED_RESPONSE))

        # The response should have a location header that is the actual
        # location of the image stored on AWS
        location = response.headers.get('location')
        if not location:
            raise AbodeException((ERROR.CAM_IMAGE_NO_LOCATION_HEADER))

        self._image_url = location

        return True

    def image_to_file(self, path, get_image=True):
        """Write the image to a file."""
        if not self.image_url or get_image:
            if not self.refresh_image():
                return False

        response = requests.get(self.image_url, stream=True)

        if response.status_code != 200:
            _LOGGER.warning(
                "Unexpected response code %s when requesting image: %s",
                str(response.status_code),
                response.text,
            )
            raise AbodeException((ERROR.CAM_IMAGE_REQUEST_INVALID))

        with open(path, 'wb') as imgfile:
            copyfileobj(response.raw, imgfile)

        return True

    def privacy_mode(self, enable):
        """Set camera privacy mode (camera on/off)."""
        if self._json_state['privacy']:
            privacy = '1' if enable else '0'

            url = CONST.PARAMS_URL + self.device_id

            camera_data = {
                'mac': self._json_state['camera_mac'],
                'privacy': privacy,
                'action': 'setParam',
                'id': self.device_id,
            }

            response = self._abode.send_request(method="put", url=url, data=camera_data)
            response_object = json.loads(response.text)

            _LOGGER.debug("Camera Privacy Mode Response: %s", response.text)

            if response_object['id'] != self.device_id:
                raise AbodeException((ERROR.SET_STATUS_DEV_ID))

            if response_object['privacy'] != str(privacy):
                raise AbodeException((ERROR.SET_PRIVACY_MODE))

            _LOGGER.info("Set camera %s privacy mode to: %s", self.device_id, privacy)

            return True

        return False

    @property
    def image_url(self):
        """Get image URL."""
        return self._image_url

    @property
    def is_on(self):
        """Get camera state (assumed on)."""
        return self.status not in (CONST.STATUS_OFF, CONST.STATUS_OFFLINE)



"""Test the Abode device classes."""
import json
import unittest

import requests_mock

import jaraco.abode
import jaraco.abode.helpers.constants as CONST

import tests.mock as MOCK
import tests.mock.login as LOGIN
import tests.mock.oauth_claims as OAUTH_CLAIMS
import tests.mock.logout as LOGOUT
import tests.mock.panel as PANEL
import tests.mock.automation as AUTOMATION
import pytest


USERNAME = 'foobar'
PASSWORD = 'deadbeef'
AID_1 = '47fae27488f74f55b964a81a066c3a01'
AID_2 = '47fae27488f74f55b964a81a066c3a02'
AID_3 = '47fae27488f74f55b964a81a066c3a03'


@requests_mock.Mocker()
class TestDevice(unittest.TestCase):
    """Test the generic AbodePy device class."""

    def setUp(self):
        """Set up Abode module."""
        self.abode = jaraco.abode.Abode(
            username=USERNAME, password=PASSWORD, disable_cache=True
        )

    def tearDown(self):
        """Clean up after test."""
        self.abode = None

    def tests_automation_init(self, m):
        """Check the Abode automation class init's properly."""
        # Set up URLs
        m.post(CONST.LOGIN_URL, text=LOGIN.post_response_ok())
        m.get(CONST.OAUTH_TOKEN_URL, text=OAUTH_CLAIMS.get_response_ok())
        m.post(CONST.LOGOUT_URL, text=LOGOUT.post_response_ok())
        m.get(CONST.PANEL_URL, text=PANEL.get_response_ok())

        # Set up automation
        automation_text = AUTOMATION.get_response_ok(
            name='Auto Away', enabled=True, aid=AID_1
        )

        automation_json = json.loads(automation_text)

        m.get(CONST.AUTOMATION_URL, text=automation_text)

        # Logout to reset everything
        self.abode.logout()

        # Get our specific automation
        automation = self.abode.get_automation(AID_1)

        # Check automation states match
        assert automation is not None
        # pylint: disable=W0212
        assert automation._automation == automation_json
        assert automation.automation_id == str(automation_json['id'])
        assert automation.name == automation_json['name']
        assert automation.is_enabled == automation_json['enabled']
        assert automation.desc is not None

    def tests_automation_refresh(self, m):
        """Check the automation Abode class refreshes."""
        # Set up URL's
        m.post(CONST.LOGIN_URL, text=LOGIN.post_response_ok())
        m.get(CONST.OAUTH_TOKEN_URL, text=OAUTH_CLAIMS.get_response_ok())
        m.post(CONST.LOGOUT_URL, text=LOGOUT.post_response_ok())
        m.get(CONST.PANEL_URL, text=PANEL.get_response_ok())

        # Set up automation
        automation_text = (
            '['
            + AUTOMATION.get_response_ok(
                name='Test Automation', enabled=True, aid=AID_1
            )
            + ']'
        )

        automation_json = json.loads(automation_text)

        m.get(CONST.AUTOMATION_URL, text=automation_text)

        # Set up refreshed automation
        automation_text_changed = (
            '['
            + AUTOMATION.get_response_ok(
                name='Test Automation Changed', enabled=False, aid=AID_1
            )
            + ']'
        )

        automation_json_changed = json.loads(automation_text_changed)

        automation_id_url = str.replace(
            CONST.AUTOMATION_ID_URL, '$AUTOMATIONID$', str(automation_json[0]['id'])
        )

        m.get(automation_id_url, text=automation_text_changed)

        # Logout to reset everything
        self.abode.logout()

        # Get the first automation and test
        automation = self.abode.get_automation(AID_1)

        # Check automation states match
        # pylint: disable=W0212
        assert automation is not None
        assert automation._automation == automation_json[0]

        # Refresh and retest
        automation.refresh()
        assert automation._automation == automation_json_changed[0]

        # Refresh with get_automation() and test
        automation_text_changed = (
            '['
            + AUTOMATION.get_response_ok(
                name='Test Automation Changed Again', enabled=True, aid=AID_1
            )
            + ']'
        )

        automation_json_changed = json.loads(automation_text_changed)
        m.get(automation_id_url, text=automation_text_changed)

        # Refresh and retest
        automation = self.abode.get_automation(AID_1, refresh=True)

        assert automation._automation == automation_json_changed[0]

        # Test refresh returning an incorrect ID throws exception
        # Set up refreshed automation
        automation_text_changed = (
            '['
            + AUTOMATION.get_response_ok(
                name='Test Automation Changed',
                enabled=False,
                aid='47fae27488f74f55b964a81a066c3a11',
            )
            + ']'
        )

        m.get(automation_id_url, text=automation_text_changed)

        with pytest.raises(jaraco.abode.AbodeException):
            automation.refresh()

    def tests_multiple_automations(self, m):
        """Check that multiple automations work and return correctly."""
        # Set up URL's
        m.post(CONST.LOGIN_URL, text=LOGIN.post_response_ok())
        m.get(CONST.OAUTH_TOKEN_URL, text=OAUTH_CLAIMS.get_response_ok())
        m.post(CONST.LOGOUT_URL, text=LOGOUT.post_response_ok())
        m.get(CONST.PANEL_URL, text=PANEL.get_response_ok())

        # Set up automations
        automation_text = (
            '['
            + AUTOMATION.get_response_ok(
                name='Test Automation One', enabled=True, aid=AID_1
            )
            + ','
            + AUTOMATION.get_response_ok(
                name='Test Automation Two', enabled=True, aid=AID_2
            )
            + ','
            + AUTOMATION.get_response_ok(
                name='Test Automation Three', enabled=True, aid=AID_3
            )
            + ']'
        )

        automation_json = json.loads(automation_text)

        m.get(CONST.AUTOMATION_URL, text=automation_text)

        # Logout to reset everything
        self.abode.logout()

        # Test that the automations return the correct number
        automations = self.abode.get_automations()
        assert len(automations) == 3

        # Get the first automation and test
        # pylint: disable=W0212
        automation_1 = self.abode.get_automation(AID_1)
        assert automation_1 is not None
        assert automation_1._automation == automation_json[0]

        automation_2 = self.abode.get_automation(AID_2)
        assert automation_2 is not None
        assert automation_2._automation == automation_json[1]

        automation_3 = self.abode.get_automation(AID_3)
        assert automation_3 is not None
        assert automation_3._automation == automation_json[2]

    def tests_automation_class_reuse(self, m):
        """Check that automations reuse the same classes when refreshed."""
        # Set up URL's
        m.post(CONST.LOGIN_URL, text=LOGIN.post_response_ok())
        m.get(CONST.OAUTH_TOKEN_URL, text=OAUTH_CLAIMS.get_response_ok())
        m.post(CONST.LOGOUT_URL, text=LOGOUT.post_response_ok())
        m.get(CONST.PANEL_URL, text=PANEL.get_response_ok())

        # Set up automations
        automation_text = (
            '['
            + AUTOMATION.get_response_ok(
                name='Test Automation One', enabled=True, aid=AID_1
            )
            + ','
            + AUTOMATION.get_response_ok(
                name='Test Automation Two', enabled=True, aid=AID_2
            )
            + ']'
        )

        automation_json = json.loads(automation_text)

        m.get(CONST.AUTOMATION_URL, text=automation_text)

        # Logout to reset everything
        self.abode.logout()

        # Test that the automations return the correct number
        automations = self.abode.get_automations()
        assert len(automations) == 2

        # Get the automations and test
        # pylint: disable=W0212
        automation_1 = self.abode.get_automation(AID_1)
        assert automation_1 is not None
        assert automation_1._automation == automation_json[0]

        automation_2 = self.abode.get_automation(AID_2)
        assert automation_2 is not None
        assert automation_2._automation == automation_json[1]

        # Update the automations
        automation_text = (
            '['
            + AUTOMATION.get_response_ok(
                name='Test Automation One Changed', enabled=False, aid=AID_1
            )
            + ','
            + AUTOMATION.get_response_ok(
                name='Test Automation Two Changed', enabled=False, aid=AID_2
            )
            + ','
            + AUTOMATION.get_response_ok(
                name='Test Automation Three New', enabled=True, aid=AID_3
            )
            + ']'
        )

        automation_json_changed = json.loads(automation_text)

        m.get(CONST.AUTOMATION_URL, text=automation_text)

        # Update
        automations_changed = self.abode.get_automations(refresh=True)
        assert len(automations_changed) == 3

        # Check that the original two automations have updated
        # and are using the same class
        automation_1_changed = self.abode.get_automation(AID_1)
        assert automation_1_changed is not None
        assert automation_1_changed._automation == automation_json_changed[0]
        assert automation_1 is automation_1_changed

        automation_2_changed = self.abode.get_automation(AID_2)
        assert automation_2_changed is not None
        assert automation_2_changed._automation == automation_json_changed[1]
        assert automation_2 is automation_2_changed

        # Check that the third new automation is correct
        automation_3 = self.abode.get_automation(AID_3)
        assert automation_3 is not None
        assert automation_3._automation == automation_json_changed[2]

    def tests_automation_enable(self, m):
        """Check that automations can change their enable state."""
        # Set up URL's
        m.post(CONST.LOGIN_URL, text=LOGIN.post_response_ok())
        m.get(CONST.OAUTH_TOKEN_URL, text=OAUTH_CLAIMS.get_response_ok())
        m.post(CONST.LOGOUT_URL, text=LOGOUT.post_response_ok())
        m.get(CONST.PANEL_URL, text=PANEL.get_response_ok())

        # Set up automation
        automation_text = (
            '['
            + AUTOMATION.get_response_ok(
                name='Test Automation One', enabled=True, aid=AID_1
            )
            + ']'
        )

        automation_json = json.loads(automation_text)

        m.get(CONST.AUTOMATION_URL, text=automation_text)

        # Logout to reset everything
        self.abode.logout()

        # Get the automation and test
        # pylint: disable=W0212
        automation = self.abode.get_automation(AID_1)
        assert automation is not None
        assert automation._automation == automation_json[0]
        assert automation.is_enabled

        # Set up our active state change and URL
        set_active_url = str.replace(
            CONST.AUTOMATION_ID_URL, '$AUTOMATIONID$', str(automation_json[0]['id'])
        )

        m.patch(
            set_active_url,
            text=AUTOMATION.get_response_ok(
                name='Test Automation One', enabled=False, aid=AID_1
            ),
        )

        # Test the changed state
        automation.enable(False)
        assert not automation.is_enabled

        # Change the state back, this time with an array response
        m.patch(
            set_active_url,
            text='['
            + AUTOMATION.get_response_ok(
                name='Test Automation One', enabled=True, aid=AID_1
            )
            + ']',
        )

        # Test the changed state
        automation.enable(True)
        assert automation.is_enabled

        # Test that the response returns the wrong state
        m.patch(
            set_active_url,
            text='['
            + AUTOMATION.get_response_ok(
                name='Test Automation One', enabled=True, aid=AID_1
            )
            + ']',
        )

        with pytest.raises(jaraco.abode.AbodeException):
            automation.enable(False)

        # Test that the response returns the wrong id
        m.patch(
            set_active_url,
            text='['
            + AUTOMATION.get_response_ok(
                name='Test Automation One', enabled=True, aid=AID_2
            )
            + ']',
        )

        with pytest.raises(jaraco.abode.AbodeException):
            automation.enable(True)

    def tests_automation_trigger(self, m):
        """Check that automations can be triggered."""
        # Set up URL's
        m.post(CONST.LOGIN_URL, text=LOGIN.post_response_ok())
        m.get(CONST.OAUTH_TOKEN_URL, text=OAUTH_CLAIMS.get_response_ok())
        m.post(CONST.LOGOUT_URL, text=LOGOUT.post_response_ok())
        m.get(CONST.PANEL_URL, text=PANEL.get_response_ok())

        # Set up automation
        automation_text = (
            '['
            + AUTOMATION.get_response_ok(
                name='Test Automation One', enabled=True, aid=AID_1
            )
            + ']'
        )

        m.get(CONST.AUTOMATION_URL, text=automation_text)

        # Logout to reset everything
        self.abode.logout()

        # Get the automation and test
        # pylint: disable=W0212
        automation = self.abode.get_automation(AID_1)
        assert automation is not None

        # Set up our automation trigger reply
        set_active_url = str.replace(
            CONST.AUTOMATION_APPLY_URL, '$AUTOMATIONID$', automation.automation_id
        )
        m.post(set_active_url, text=MOCK.generic_response_ok())

        # Test triggering
        assert automation.trigger()



"""Mock Abode Login Response."""

from tests.mock import AUTH_TOKEN
import tests.mock.panel as panel
import tests.mock.user as user


def post_response_ok(auth_token=AUTH_TOKEN, user_response=user.get_response_ok()):
    """Return the successful login response json."""
    return (
        '''
    {
       "token":"'''
        + auth_token
        + '''",
       "expired_at":"2017-06-05 00:14:12",
       "initiate_screen":"timeline",
       "user":'''
        + user_response
        + ''',
       "panel":'''
        + panel.get_response_ok()
        + ''',
       "permissions":{
          "premium_streaming":"0",
          "guest_app":"0",
          "family_app":"0",
          "multiple_accounts":"1",
          "google_voice":"1",
          "nest":"1",
          "alexa":"1",
          "ifttt":"1",
          "no_associates":"100",
          "no_contacts":"2",
          "no_devices":"155",
          "no_ipcam":"100",
          "no_quick_action":"25",
          "no_automation":"75",
          "media_storage":"3",
          "cellular_backup":"0",
          "cms_duration":"",
          "cms_included":"0"
       },
       "integrations":{
          "nest":{
             "is_connected":0,
             "is_home_selected":0
          }
       }
    }'''
    )


def post_response_bad_request():
    """Return the failed login response json."""
    return '''
        {
            "code":400,"message":"Username and password do not match.",
            "detail":null
        }'''


def post_response_mfa_code_required():
    """Return the MFA code required login response json."""
    return '''
        {
            "code":200,"mfa_type":"google_authenticator",
            "detail":null
        }'''


def post_response_bad_mfa_code():
    """Return the bad MFA code login response json."""
    return '''
        {
            "code":400,"message":"Invalid authentication key.",
            "detail":null
        }'''


def post_response_unknown_mfa_type():
    """Return a login response json with an unknown mfa type."""
    return '''
        {
            "code":200,"mfa_type":"sms",
            "detail":null
        }'''



"""Abode sensor device."""
import re

from ..devices.binary_sensor import AbodeBinarySensor
from ..helpers import constants as CONST


class AbodeSensor(AbodeBinarySensor):
    """Class to represent a sensor device."""

    def _get_status(self, key):
        return self._json_state.get(CONST.STATUSES_KEY, {}).get(key)

    def _get_numeric_status(self, key):
        """Extract the numeric value from the statuses object."""
        value = self._get_status(key)

        if value and any(i.isdigit() for i in value):
            return float(re.sub("[^0-9.]", "", value))

    @property
    def temp(self):
        """Get device temp."""
        return self._get_numeric_status(CONST.TEMP_STATUS_KEY)

    @property
    def temp_unit(self):
        """Get unit of temp."""
        if CONST.UNIT_FAHRENHEIT in self._get_status(CONST.TEMP_STATUS_KEY):
            return CONST.UNIT_FAHRENHEIT

        if CONST.UNIT_CELSIUS in self._get_status(CONST.TEMP_STATUS_KEY):
            return CONST.UNIT_CELSIUS

    @property
    def humidity(self):
        """Get device humdity."""
        return self._get_numeric_status(CONST.HUMI_STATUS_KEY)

    @property
    def humidity_unit(self):
        """Get unit of humidity."""
        if CONST.UNIT_PERCENT in self._get_status(CONST.HUMI_STATUS_KEY):
            return CONST.UNIT_PERCENT

    @property
    def lux(self):
        """Get device lux."""
        return self._get_numeric_status(CONST.LUX_STATUS_KEY)

    @property
    def lux_unit(self):
        """Get unit of lux."""
        if CONST.UNIT_LUX in self._get_status(CONST.LUX_STATUS_KEY):
            return CONST.LUX

    @property
    def has_temp(self):
        """Device reports temperature."""
        return self.temp is not None

    @property
    def has_humidity(self):
        """Device reports humidity level."""
        return self.humidity is not None

    @property
    def has_lux(self):
        """Device reports light lux level."""
        return self.lux is not None



"""Test the Abode device classes."""
import unittest

import requests_mock

import jaraco.abode
import jaraco.abode.helpers.constants as CONST

import tests.mock.login as LOGIN
import tests.mock.oauth_claims as OAUTH_CLAIMS
import tests.mock.logout as LOGOUT
import tests.mock.panel as PANEL
import tests.mock.devices as DEVICES
import tests.mock.devices.alarm as ALARM
import pytest


USERNAME = 'foobar'
PASSWORD = 'deadbeef'


class TestAlarm(unittest.TestCase):
    """Test the generic AbodePy device class."""

    def setUp(self):
        """Set up Abode module."""
        self.abode = jaraco.abode.Abode(
            username=USERNAME, password=PASSWORD, disable_cache=True
        )

    def tearDown(self):
        """Clean up after test."""
        self.abode = None

    @requests_mock.mock()
    def tests_abode_alarm_setup(self, m):
        """Check that Abode alarm device is set up properly."""
        panel = PANEL.get_response_ok(mode=CONST.MODE_STANDBY)
        alarm = ALARM.device(area='1', panel=panel)

        m.post(CONST.LOGIN_URL, text=LOGIN.post_response_ok())
        m.get(CONST.OAUTH_TOKEN_URL, text=OAUTH_CLAIMS.get_response_ok())
        m.get(CONST.DEVICES_URL, text=DEVICES.EMPTY_DEVICE_RESPONSE)
        m.get(CONST.PANEL_URL, text=PANEL.get_response_ok())

        alarm_device = self.abode.get_alarm()

        assert alarm_device is not None
        # pylint: disable=W0212
        assert alarm_device._json_state == alarm

    @requests_mock.mock()
    def tests_alarm_device_properties(self, m):
        """Check that the abode device properties are working."""
        # Set up URL's
        m.post(CONST.LOGIN_URL, text=LOGIN.post_response_ok())
        m.get(CONST.OAUTH_TOKEN_URL, text=OAUTH_CLAIMS.get_response_ok())
        m.post(CONST.LOGOUT_URL, text=LOGOUT.post_response_ok())
        m.get(
            CONST.PANEL_URL,
            text=PANEL.get_response_ok(
                mode=CONST.MODE_STANDBY,
                battery=True,
                is_cellular=True,
                mac='01:AA:b3:C4:d5:66',
            ),
        )
        m.get(CONST.DEVICES_URL, text=DEVICES.EMPTY_DEVICE_RESPONSE)

        # Logout to reset everything
        self.abode.logout()

        # Get alarm and test
        alarm = self.abode.get_alarm()
        assert alarm is not None
        assert alarm.mode == CONST.MODE_STANDBY
        assert alarm.status == CONST.MODE_STANDBY
        assert alarm.battery
        assert alarm.is_cellular
        assert not alarm.is_on
        assert alarm.device_uuid == '01aab3c4d566'
        assert alarm.mac_address == '01:AA:b3:C4:d5:66'

        # Change alarm properties and state to away and test
        m.get(
            CONST.PANEL_URL,
            text=PANEL.get_response_ok(
                mode=CONST.MODE_AWAY, battery=False, is_cellular=False
            ),
        )

        # Refresh alarm and test
        alarm.refresh()

        assert alarm.mode == CONST.MODE_AWAY
        assert alarm.status == CONST.MODE_AWAY
        assert not alarm.battery
        assert not alarm.is_cellular
        assert alarm.is_on

        # Change alarm state to final on state and test
        m.get(CONST.PANEL_URL, text=PANEL.get_response_ok(mode=CONST.MODE_HOME))

        # Refresh alarm and test
        alarm.refresh()
        assert alarm.mode == CONST.MODE_HOME
        assert alarm.status == CONST.MODE_HOME
        assert alarm.is_on

    @requests_mock.mock()
    def tests_alarm_device_mode_changes(self, m):
        """Test that the abode alarm can change/report modes."""
        # Set up URL's
        m.post(CONST.LOGIN_URL, text=LOGIN.post_response_ok())
        m.get(CONST.OAUTH_TOKEN_URL, text=OAUTH_CLAIMS.get_response_ok())
        m.post(CONST.LOGOUT_URL, text=LOGOUT.post_response_ok())
        m.get(CONST.PANEL_URL, text=PANEL.get_response_ok(mode=CONST.MODE_STANDBY))
        m.get(CONST.DEVICES_URL, text=DEVICES.EMPTY_DEVICE_RESPONSE)

        # Logout to reset everything
        self.abode.logout()

        # Assert that after login we have our alarm device with standby mode
        alarm = self.abode.get_alarm()

        assert alarm is not None
        assert alarm.status == CONST.MODE_STANDBY

        # Set mode URLs
        m.put(
            CONST.get_panel_mode_url('1', CONST.MODE_STANDBY),
            text=PANEL.put_response_ok(mode=CONST.MODE_STANDBY),
        )
        m.put(
            CONST.get_panel_mode_url('1', CONST.MODE_AWAY),
            text=PANEL.put_response_ok(mode=CONST.MODE_AWAY),
        )
        m.put(
            CONST.get_panel_mode_url('1', CONST.MODE_HOME),
            text=PANEL.put_response_ok(mode=CONST.MODE_HOME),
        )

        # Set and test text based mode changes
        assert alarm.set_mode(CONST.MODE_HOME)
        assert alarm.mode == CONST.MODE_HOME
        assert not alarm.is_standby
        assert alarm.is_home
        assert not alarm.is_away

        assert alarm.set_mode(CONST.MODE_AWAY)
        assert alarm.mode == CONST.MODE_AWAY
        assert not alarm.is_standby
        assert not alarm.is_home
        assert alarm.is_away

        assert alarm.set_mode(CONST.MODE_STANDBY)
        assert alarm.mode == CONST.MODE_STANDBY
        assert alarm.is_standby
        assert not alarm.is_home
        assert not alarm.is_away

        # Set and test direct mode changes
        assert alarm.set_home()
        assert alarm.mode == CONST.MODE_HOME
        assert not alarm.is_standby
        assert alarm.is_home
        assert not alarm.is_away

        assert alarm.set_away()
        assert alarm.mode == CONST.MODE_AWAY
        assert not alarm.is_standby
        assert not alarm.is_home
        assert alarm.is_away

        assert alarm.set_standby()
        assert alarm.mode == CONST.MODE_STANDBY
        assert alarm.is_standby
        assert not alarm.is_home
        assert not alarm.is_away

        # Set and test default mode changes
        assert alarm.switch_off()
        assert alarm.mode == CONST.MODE_STANDBY
        assert alarm.is_standby
        assert not alarm.is_home
        assert not alarm.is_away

        self.abode.set_default_mode(CONST.MODE_HOME)
        assert alarm.switch_on()
        assert alarm.mode == CONST.MODE_HOME
        assert not alarm.is_standby
        assert alarm.is_home
        assert not alarm.is_away

        assert alarm.switch_off()
        assert alarm.mode == CONST.MODE_STANDBY
        assert alarm.is_standby
        assert not alarm.is_home
        assert not alarm.is_away

        self.abode.set_default_mode(CONST.MODE_AWAY)
        assert alarm.switch_on()
        assert alarm.mode == CONST.MODE_AWAY
        assert not alarm.is_standby
        assert not alarm.is_home
        assert alarm.is_away

        # Test that no mode throws exception
        with pytest.raises(jaraco.abode.AbodeException):
            alarm.set_mode(mode=None)

        # Test that an invalid mode throws exception
        with pytest.raises(jaraco.abode.AbodeException):
            alarm.set_mode('chestnuts')

        # Test that an invalid mode change response throws exception
        m.put(
            CONST.get_panel_mode_url('1', CONST.MODE_HOME),
            text=PANEL.put_response_ok(mode=CONST.MODE_AWAY),
        )

        with pytest.raises(jaraco.abode.AbodeException):
            alarm.set_mode(CONST.MODE_HOME)

        # Test that an invalid area in mode change response throws exception
        m.put(
            CONST.get_panel_mode_url('1', CONST.MODE_HOME),
            text=PANEL.put_response_ok(area='2', mode=CONST.MODE_HOME),
        )

        with pytest.raises(jaraco.abode.AbodeException):
            alarm.set_mode(CONST.MODE_HOME)



"""Abodepy utility methods."""
import logging
import pickle
import uuid

_LOGGER = logging.getLogger(__name__)


def save_cache(data, filename):
    """Save cookies to a file."""
    with open(filename, 'wb') as handle:
        pickle.dump(data, handle)


def load_cache(filename):
    """Load cookies from a file."""
    with open(filename, 'rb') as handle:
        try:
            return pickle.load(handle)
        except EOFError:
            _LOGGER.warning("Empty pickle file: %s", filename)
        except (pickle.UnpicklingError, ValueError):
            _LOGGER.warning("Corrupted pickle file: %s", filename)


def gen_uuid():
    """Generate a new Abode UUID."""
    return str(uuid.uuid1())


def update(dct, dct_merge):
    """Recursively merge dicts."""
    for key, value in dct_merge.items():
        if key in dct and isinstance(dct[key], dict):
            dct[key] = update(dct[key], value)
        else:
            dct[key] = value
    return dct



"""
Mock responses that mimic actual data from Abode servers.

This file should be updated any time the Abode server responses
change to confirm that this library can still communicate.
"""

AUTH_TOKEN = 'web-1eb04ba2236d85f49d4b9b4bb91665f2'
OAUTH_TOKEN = 'ohyeahthisisanoauthtoken'


def response_forbidden():
    """Return the invalid API key response json."""
    return '{"code":403,"message":"Invalid API Key"}'


def generic_response_ok():
    """
    Return the successful generic change response json.

    Used for settings changes.
    """
    return '{"code":200,"message":"OK"}'



"""Mock Abode Door Lock Device."""
import jaraco.abode.helpers.constants as CONST

DEVICE_ID = 'ZW:0000006'
CONTROL_URL = 'api/v1/control/lock/' + DEVICE_ID


def device(
    devid=DEVICE_ID,
    status=CONST.STATUS_LOCKCLOSED,
    low_battery=False,
    no_response=False,
):
    """Door lock mock device."""
    return (
        '''
        {
          "id":"'''
        + devid
        + '''",
          "type_tag":"device_type.door_lock",
          "type":"Door Lock",
          "name":"Back Door Deadbolt",
          "area":"1",
          "zone":"7",
          "sort_order":"0",
          "is_window":"0",
          "bypass":"0",
          "schar_24hr":"0",
          "sresp_mode_0":"0",
          "sresp_entry_0":"0",
          "sresp_exit_0":"0",
          "sresp_mode_1":"0",
          "sresp_entry_1":"0",
          "sresp_exit_1":"0",
          "sresp_mode_2":"0",
          "sresp_entry_2":"0",
          "sresp_exit_2":"0",
          "sresp_mode_3":"0",
          "sresp_entry_3":"0",
          "sresp_exit_3":"0",
          "version":"",
          "origin":"abode",
          "control_url":"'''
        + CONTROL_URL
        + '''",
          "deep_link":null,
          "status_color":"#5cb85c",
          "faults":{
             "low_battery":'''
        + str(int(low_battery))
        + ''',
             "tempered":0,
             "supervision":0,
             "out_of_order":0,
             "no_response":'''
        + str(int(no_response))
        + '''
          },
          "status":"'''
        + status
        + '''",
          "statuses":{
             "hvac_mode":null
          },
          "status_ex":"",
          "actions":[
             {
                "label":"Lock",
                "value":"a=1&z=7&sw=on;"
             },
             {
                "label":"Unlock",
                "value":"a=1&z=7&sw=off;"
             }
          ],
          "status_icons":{
             "LockOpen":"assets/icons/unlocked-red.svg",
             "LockClosed":"assets/icons/locked-green.svg"
          },
          "automation_settings":null,
          "icon":"assets/icons/automation-lock.svg"
       }'''
    )



"""Abode light device."""
import json
import logging
import math

from ..exceptions import AbodeException

from ..devices.switch import AbodeSwitch
from ..helpers import constants as CONST
from ..helpers import errors as ERROR


_LOGGER = logging.getLogger(__name__)


class AbodeLight(AbodeSwitch):
    """Class for lights (dimmers)."""

    def set_color_temp(self, color_temp):
        """Set device color."""
        if self._json_state['control_url']:
            url = CONST.INTEGRATIONS_URL + self._device_uuid

            color_data = {
                'action': 'setcolortemperature',
                'colorTemperature': int(color_temp),
            }

            response = self._abode.send_request("post", url, data=color_data)
            response_object = json.loads(response.text)

            _LOGGER.debug("Set Color Temp Response: %s", response.text)

            if response_object['idForPanel'] != self.device_id:
                raise AbodeException((ERROR.SET_STATUS_DEV_ID))

            if response_object['colorTemperature'] != int(color_temp):
                _LOGGER.warning(
                    (
                        "Set color temp mismatch for device %s. "
                        "Request val: %s, Response val: %s "
                    ),
                    self.device_id,
                    color_temp,
                    response_object['colorTemperature'],
                )

                color_temp = response_object['colorTemperature']

            self.update({'statuses': {'color_temp': color_temp}})

            _LOGGER.info("Set device %s color_temp to: %s", self.device_id, color_temp)
            return True

        return False

    def set_color(self, color):
        """Set device color."""
        if self._json_state['control_url']:
            url = CONST.INTEGRATIONS_URL + self._device_uuid

            hue, saturation = color
            color_data = {
                'action': 'setcolor',
                'hue': int(hue),
                'saturation': int(saturation),
            }

            response = self._abode.send_request("post", url, data=color_data)
            response_object = json.loads(response.text)

            _LOGGER.debug("Set Color Response: %s", response.text)

            if response_object['idForPanel'] != self.device_id:
                raise AbodeException((ERROR.SET_STATUS_DEV_ID))

            # Abode will sometimes return hue value off by 1 (rounding error)
            hue_comparison = math.isclose(response_object["hue"], int(hue), abs_tol=1)
            if not hue_comparison or (response_object["saturation"] != int(saturation)):
                _LOGGER.warning(
                    (
                        "Set color mismatch for device %s. "
                        "Request val: %s, Response val: %s "
                    ),
                    self.device_id,
                    (hue, saturation),
                    (response_object['hue'], response_object['saturation']),
                )

                hue = response_object['hue']
                saturation = response_object['saturation']

            self.update({'statuses': {'hue': hue, 'saturation': saturation}})

            _LOGGER.info(
                "Set device %s color to: %s", self.device_id, (hue, saturation)
            )

            return True

        return False

    @property
    def brightness(self):
        """Get light brightness."""
        return self.get_value(CONST.STATUSES_KEY).get('level')

    @property
    def color_temp(self):
        """Get light color temp."""
        return self.get_value(CONST.STATUSES_KEY).get('color_temp')

    @property
    def color(self):
        """Get light color."""
        return (
            self.get_value(CONST.STATUSES_KEY).get('hue'),
            self.get_value(CONST.STATUSES_KEY).get('saturation'),
        )

    @property
    def has_brightness(self):
        """Device has brightness."""
        return self.brightness

    @property
    def has_color(self):
        """Device is using color mode."""
        if self.get_value(CONST.STATUSES_KEY).get('color_mode') == str(
            CONST.COLOR_MODE_ON
        ):
            return True
        return False

    @property
    def is_color_capable(self):
        """Device is color compatible."""
        return 'RGB' in self._type

    @property
    def is_dimmable(self):
        """Device is dimmable."""
        return 'Dimmer' in self._type



"""Representation of an automation configured in Abode."""
import json
import logging

from .exceptions import AbodeException

from .helpers import constants as CONST
from .helpers import errors as ERROR

_LOGGER = logging.getLogger(__name__)


class AbodeAutomation:
    """Class for viewing and controlling automations."""

    def __init__(self, abode, automation):
        """Init AbodeAutomation class."""
        self._abode = abode
        self._automation = automation

    def enable(self, enable):
        """Enable or disable the automation."""
        url = str.replace(CONST.AUTOMATION_ID_URL, '$AUTOMATIONID$', self.automation_id)

        self._automation['enabled'] = enable

        response = self._abode.send_request(
            method="patch", url=url, data={'enabled': enable}
        )

        response_object = json.loads(response.text)

        if isinstance(response_object, (tuple, list)):
            response_object = response_object[0]

        if str(response_object['id']) != str(self._automation['id']) or str(
            response_object['enabled']
        ) != str(self._automation['enabled']):
            raise AbodeException((ERROR.INVALID_AUTOMATION_EDIT_RESPONSE))

        self.update(response_object)

        _LOGGER.info("Set automation %s enable to: %s", self.name, self.is_enabled)
        _LOGGER.debug("Automation response: %s", response.text)

        return True

    def trigger(self):
        """Trigger the automation."""
        url = str.replace(
            CONST.AUTOMATION_APPLY_URL, '$AUTOMATIONID$', self.automation_id
        )

        self._abode.send_request(method="post", url=url)

        _LOGGER.info("Automation triggered: %s", self.name)

        return True

    def refresh(self):
        """Refresh the automation."""
        url = str.replace(CONST.AUTOMATION_ID_URL, '$AUTOMATIONID$', self.automation_id)

        response = self._abode.send_request(method="get", url=url)
        response_object = json.loads(response.text)

        if isinstance(response_object, (tuple, list)):
            response_object = response_object[0]

        if str(response_object['id']) != self.automation_id:
            raise AbodeException((ERROR.INVALID_AUTOMATION_REFRESH_RESPONSE))

        self.update(response_object)

    def update(self, automation):
        """Update the internal automation json."""
        self._automation.update(
            {k: automation[k] for k in automation if self._automation.get(k)}
        )

    @property
    def automation_id(self):
        """Get the id of the automation."""
        return str(self._automation['id'])

    @property
    def name(self):
        """Get the name of the automation."""
        return self._automation['name']

    @property
    def is_enabled(self):
        """Return True if the automation is enabled."""
        return self._automation['enabled']

    @property
    def desc(self):
        """Get a short description of the automation."""
        return '{0} (ID: {1}, Enabled: {2})'.format(
            self.name, self.automation_id, self.is_enabled
        )



"""Mock Abode Power Switch Sensor Device."""
import jaraco.abode.helpers.constants as CONST

DEVICE_ID = 'ZW:0000000a'
CONTROL_URL = 'api/v1/control/power_switch/' + DEVICE_ID


def device(
    devid=DEVICE_ID, status=CONST.STATUS_OPEN, low_battery=False, no_response=False
):
    """Secure barrier mock device."""
    return (
        '''
        {
           "id":"'''
        + devid
        + '''",
           "type_tag":"device_type.secure_barrier",
           "type":"Secure Barrier",
           "name":"Garage Auto Door",
           "area":"1",
           "zone":"11",
           "sort_order":"0",
           "is_window":"0",
           "bypass":"0",
           "schar_24hr":"0",
           "sresp_mode_0":"0",
           "sresp_entry_0":"0",
           "sresp_exit_0":"0",
           "sresp_mode_1":"0",
           "sresp_entry_1":"0",
           "sresp_exit_1":"0",
           "sresp_mode_2":"0",
           "sresp_entry_2":"0",
           "sresp_exit_2":"0",
           "sresp_mode_3":"0",
           "sresp_entry_3":"0",
           "sresp_exit_3":"0",
           "capture_mode":null,
           "origin":"abode",
           "control_url":"'''
        + CONTROL_URL
        + '''",
           "deep_link":null,
           "status_color":"#5cb85c",
           "faults":{
              "low_battery":'''
        + str(int(low_battery))
        + ''',
              "tempered":0,
              "supervision":0,
              "out_of_order":0,
              "no_response":'''
        + str(int(no_response))
        + '''
           },
           "status":"'''
        + status
        + '''",
           "statuses":{
              "hvac_mode":null
           },
           "status_ex":"",
           "actions":[
              {
                 "label":"Close",
                 "value":"a=1&z=11&sw=off;"
              },
              {
                 "label":"Open",
                 "value":"a=1&z=11&sw=on;"
              }
           ],
           "status_icons":{
              "Open":"assets/icons/garage-door-red.svg",
              "Closed":"assets/icons/garage-door-green.svg"
           },
           "icon":"assets/icons/garage-door.svg"
        }'''
    )



"""Test the Abode event controller class."""
import json
import unittest
from unittest.mock import call, Mock

import requests_mock

import jaraco.abode
import jaraco.abode.helpers.constants as CONST
import jaraco.abode.helpers.timeline as TIMELINE
from jaraco.abode.devices.binary_sensor import AbodeBinarySensor

import tests.mock.login as LOGIN
import tests.mock.oauth_claims as OAUTH_CLAIMS
import tests.mock.logout as LOGOUT
import tests.mock.panel as PANEL
import tests.mock.devices.secure_barrier as COVER
import tests.mock.devices.door_contact as DOORCONTACT
import tests.mock.devices.ir_camera as IRCAMERA
import pytest


USERNAME = 'foobar'
PASSWORD = 'deadbeef'


class TestEventController(unittest.TestCase):
    """Test the AbodePy event controller."""

    def setUp(self):
        """Set up Abode module."""
        self.abode = jaraco.abode.Abode(
            username=USERNAME, password=PASSWORD, disable_cache=True
        )

    def tearDown(self):
        """Clean up after test."""
        self.abode = None

    @requests_mock.mock()
    def tests_device_id_registration(self, m):
        """Tests that we can register for device events with a device id."""
        # Set up URL's
        m.post(CONST.LOGIN_URL, text=LOGIN.post_response_ok())
        m.get(CONST.OAUTH_TOKEN_URL, text=OAUTH_CLAIMS.get_response_ok())
        m.post(CONST.LOGOUT_URL, text=LOGOUT.post_response_ok())
        m.get(CONST.PANEL_URL, text=PANEL.get_response_ok(mode=CONST.MODE_STANDBY))
        m.get(
            CONST.DEVICES_URL,
            text=COVER.device(
                devid=COVER.DEVICE_ID,
                status=CONST.STATUS_CLOSED,
                low_battery=False,
                no_response=False,
            ),
        )

        # Logout to reset everything
        self.abode.logout()

        # Get our device
        device = self.abode.get_device(COVER.DEVICE_ID)
        assert device is not None

        # Get the event controller
        events = self.abode.events
        assert events is not None

        # Create mock callback
        callback = Mock()

        # Register our device id
        assert events.add_device_callback(device.device_id, callback)

    @requests_mock.mock()
    def tests_device_registration(self, m):
        """Tests that we can register for device events with a device."""
        # Set up URL's
        m.post(CONST.LOGIN_URL, text=LOGIN.post_response_ok())
        m.get(CONST.OAUTH_TOKEN_URL, text=OAUTH_CLAIMS.get_response_ok())
        m.post(CONST.LOGOUT_URL, text=LOGOUT.post_response_ok())
        m.get(CONST.PANEL_URL, text=PANEL.get_response_ok(mode=CONST.MODE_STANDBY))
        m.get(
            CONST.DEVICES_URL,
            text=COVER.device(
                devid=COVER.DEVICE_ID,
                status=CONST.STATUS_CLOSED,
                low_battery=False,
                no_response=False,
            ),
        )

        # Logout to reset everything
        self.abode.logout()

        # Get our device
        device = self.abode.get_device(COVER.DEVICE_ID)
        assert device is not None

        # Get the event controller
        events = self.abode.events
        assert events is not None

        def _our_callback(device):
            assert device is not None

        # Register our device
        assert events.add_device_callback(device, _our_callback)

    @requests_mock.mock()
    def tests_device_all_unregistration(self, m):
        """Tests that we can unregister for all device events."""
        # Set up URL's
        m.post(CONST.LOGIN_URL, text=LOGIN.post_response_ok())
        m.get(CONST.OAUTH_TOKEN_URL, text=OAUTH_CLAIMS.get_response_ok())
        m.post(CONST.LOGOUT_URL, text=LOGOUT.post_response_ok())
        m.get(CONST.PANEL_URL, text=PANEL.get_response_ok(mode=CONST.MODE_STANDBY))
        m.get(
            CONST.DEVICES_URL,
            text=COVER.device(
                devid=COVER.DEVICE_ID,
                status=CONST.STATUS_CLOSED,
                low_battery=False,
                no_response=False,
            ),
        )

        # Logout to reset everything
        self.abode.logout()

        # Get our device
        device = self.abode.get_device(COVER.DEVICE_ID)
        assert device is not None

        # Get the event controller
        events = self.abode.events
        assert events is not None

        def _our_callback(device):
            assert device is not None

        # Register our device
        assert events.add_device_callback(device, _our_callback)

        # Unregister all callbacks
        assert events.remove_all_device_callbacks(device)

    @requests_mock.mock()
    def tests_invalid_device(self, m):
        """Tests that invalid devices are not registered."""
        # Set up URL's
        m.post(CONST.LOGIN_URL, text=LOGIN.post_response_ok())
        m.get(CONST.OAUTH_TOKEN_URL, text=OAUTH_CLAIMS.get_response_ok())
        m.post(CONST.LOGOUT_URL, text=LOGOUT.post_response_ok())
        m.get(CONST.PANEL_URL, text=PANEL.get_response_ok(mode=CONST.MODE_STANDBY))
        m.get(
            CONST.DEVICES_URL,
            text=COVER.device(
                devid=COVER.DEVICE_ID,
                status=CONST.STATUS_CLOSED,
                low_battery=False,
                no_response=False,
            ),
        )

        # Logout to reset everything
        self.abode.logout()

        # Get our device
        device = self.abode.get_device(COVER.DEVICE_ID)
        assert device is not None

        # Get the event controller
        events = self.abode.events
        assert events is not None

        # Create mock callback
        callback = Mock()

        # Test that no device returns false
        assert not events.add_device_callback(None, callback)

        # Create a fake device and attempt to register that
        fake_device = AbodeBinarySensor(json.loads(DOORCONTACT.device()), self.abode)

        with pytest.raises(jaraco.abode.AbodeException):
            events.add_device_callback(fake_device, callback)

    @requests_mock.mock()
    def tests_invalid_all_device_unregister(self, m):
        """Tests that invalid devices are not all unregistered."""
        # Set up URL's
        m.post(CONST.LOGIN_URL, text=LOGIN.post_response_ok())
        m.get(CONST.OAUTH_TOKEN_URL, text=OAUTH_CLAIMS.get_response_ok())
        m.post(CONST.LOGOUT_URL, text=LOGOUT.post_response_ok())
        m.get(CONST.PANEL_URL, text=PANEL.get_response_ok(mode=CONST.MODE_STANDBY))
        m.get(
            CONST.DEVICES_URL,
            text=COVER.device(
                devid=COVER.DEVICE_ID,
                status=CONST.STATUS_CLOSED,
                low_battery=False,
                no_response=False,
            ),
        )

        # Logout to reset everything
        self.abode.logout()

        # Get our device
        device = self.abode.get_device(COVER.DEVICE_ID)
        assert device is not None

        # Get the event controller
        events = self.abode.events
        assert events is not None

        # Test that no device returns false
        assert not events.remove_all_device_callbacks(None)

        # Create a fake device and attempt to unregister that
        fake_device = AbodeBinarySensor(json.loads(DOORCONTACT.device()), self.abode)

        with pytest.raises(jaraco.abode.AbodeException):
            events.remove_all_device_callbacks(fake_device)

    def tests_event_registration(self):
        """Tests that events register correctly."""
        # Get the event controller
        events = self.abode.events
        assert events is not None

        # Create mock callback
        callback = Mock()

        # Test that a valid event registers
        assert events.add_event_callback(TIMELINE.ALARM_GROUP, callback)

        # Test that no event group returns false
        assert not events.add_event_callback(None, callback)

        # Test that an invalid event throws exception
        with pytest.raises(jaraco.abode.AbodeException):
            events.add_event_callback("lol", callback)

    def tests_timeline_registration(self):
        """Tests that timeline events register correctly."""
        # Get the event controller
        events = self.abode.events
        assert events is not None

        # Create mock callback
        callback = Mock()

        # Test that a valid timeline event registers
        assert events.add_timeline_callback(TIMELINE.CAPTURE_IMAGE, callback)

        # Test that no timeline event returns false
        assert not events.add_timeline_callback(None, callback)

        # Test that an invalid timeline event string throws exception
        with pytest.raises(jaraco.abode.AbodeException):
            events.add_timeline_callback("lol", callback)

        # Test that an invalid timeline event dict throws exception
        with pytest.raises(jaraco.abode.AbodeException):
            events.add_timeline_callback({"lol": "lol"}, callback)

    @requests_mock.mock()
    def tests_device_callback(self, m):
        """Tests that device updates callback correctly."""
        # Set up URL's
        m.post(CONST.LOGIN_URL, text=LOGIN.post_response_ok())
        m.get(CONST.OAUTH_TOKEN_URL, text=OAUTH_CLAIMS.get_response_ok())
        m.post(CONST.LOGOUT_URL, text=LOGOUT.post_response_ok())
        m.get(CONST.PANEL_URL, text=PANEL.get_response_ok(mode=CONST.MODE_STANDBY))
        m.get(
            CONST.DEVICES_URL,
            text=COVER.device(
                devid=COVER.DEVICE_ID,
                status=CONST.STATUS_CLOSED,
                low_battery=False,
                no_response=False,
            ),
        )

        # Logout to reset everything
        self.abode.logout()

        # Get our device
        device = self.abode.get_device(COVER.DEVICE_ID)
        assert device is not None

        # Get the event controller
        events = self.abode.events
        assert events is not None

        callback = Mock()

        # Register our device id
        assert events.add_device_callback(device.device_id, callback)

        # Set up device update URL
        device_url = str.replace(CONST.DEVICE_URL, '$DEVID$', COVER.DEVICE_ID)
        m.get(
            device_url,
            text=COVER.device(
                devid=COVER.DEVICE_ID,
                status=CONST.STATUS_OPEN,
                low_battery=False,
                no_response=False,
            ),
        )

        # Call our device callback method
        # pylint: disable=protected-access
        events._on_device_update(device.device_id)
        callback.assert_called_with(device)

        # Test that our device updated
        assert device.status == CONST.STATUS_OPEN

        # Test that no device ID cleanly returns
        events._on_device_update(None)

        # Test that an unknown device cleanly returns
        events._on_device_update(DOORCONTACT.DEVICE_ID)

    def tests_events_callback(self):
        """Tests that event updates callback correctly."""
        # Get the event controller
        events = self.abode.events
        assert events is not None

        # Create mock callbacks
        capture_callback = Mock()
        alarm_callback = Mock()

        # Register our events
        assert events.add_event_callback(TIMELINE.CAPTURE_GROUP, capture_callback)

        assert events.add_event_callback(TIMELINE.ALARM_GROUP, alarm_callback)

        # Call our events callback method and trigger a capture group event
        # pylint: disable=protected-access
        event_json = json.loads(IRCAMERA.timeline_event())
        events._on_timeline_update(event_json)

        # Our capture callback should get one, but our alarm should not
        capture_callback.assert_called_with(event_json)
        alarm_callback.assert_not_called()

        # Test that an invalid event exits cleanly
        events._on_timeline_update({"invalid": "event"})

    def tests_timeline_callback(self):
        """Tests that timeline updates callback correctly."""
        # Get the event controller
        events = self.abode.events
        assert events is not None

        # Create mock callbacks
        all_callback = Mock()
        image_callback = Mock()
        opened_callback = Mock()

        # Register our events
        assert events.add_timeline_callback(TIMELINE.ALL, all_callback)

        assert events.add_timeline_callback(TIMELINE.CAPTURE_IMAGE, image_callback)

        assert events.add_timeline_callback(TIMELINE.OPENED, opened_callback)

        # Call our events callback method and trigger an image capture event
        # pylint: disable=protected-access
        event_json = json.loads(IRCAMERA.timeline_event())
        events._on_timeline_update(event_json)

        # all and image callbacks should have one, opened none
        all_callback.assert_called_with(event_json)
        image_callback.assert_called_with(event_json)
        opened_callback.assert_not_called()

        # Test that an invalid event exits cleanly
        events._on_timeline_update({"invalid": "event"})

    @requests_mock.mock()
    def tests_alarm_callback(self, m):
        """Tests that alarm device updates callback correctly."""
        # Set up URL's
        m.post(CONST.LOGIN_URL, text=LOGIN.post_response_ok())
        m.get(CONST.OAUTH_TOKEN_URL, text=OAUTH_CLAIMS.get_response_ok())
        m.post(CONST.LOGOUT_URL, text=LOGOUT.post_response_ok())
        m.get(CONST.PANEL_URL, text=PANEL.get_response_ok(mode=CONST.MODE_STANDBY))
        m.get(
            CONST.DEVICES_URL,
            text=COVER.device(
                devid=COVER.DEVICE_ID,
                status=CONST.STATUS_CLOSED,
                low_battery=False,
                no_response=False,
            ),
        )

        # Logout to reset everything
        self.abode.logout()

        # Get our alarm device
        alarm = self.abode.get_alarm()
        assert alarm is not None

        # Get the event controller
        events = self.abode.events
        assert events is not None

        callback = Mock()

        # Register our alarm for callback
        assert events.add_device_callback(alarm.device_id, callback)

        # Call our mode changed callback method
        # pylint: disable=protected-access
        events._on_mode_change(CONST.MODE_HOME)
        callback.assert_called_with(alarm)

        # Test that our alarm state is set properly
        assert alarm.mode == CONST.MODE_HOME

        # Test that no mode cleanly returns
        events._on_mode_change(None)

        # Test that an unknown mode cleanly returns
        events._on_mode_change("lol")

    def tests_execute_callback(self):
        """Tests that callbacks that throw exceptions don't bomb."""
        # Get the event controller
        events = self.abode.events
        assert events is not None

        # Create callbacks
        def _callback(event_json):
            raise Exception("CHAOS!!!")

        # Register events callback
        assert events.add_timeline_callback(TIMELINE.CAPTURE_IMAGE, _callback)

        # Call our events callback method and trigger an image capture event
        # pylint: disable=protected-access
        event_json = json.loads(IRCAMERA.timeline_event())
        events._on_timeline_update(event_json)

    @requests_mock.mock()
    def tests_multi_device_callback(self, m):
        """Tests that multiple device updates callback correctly."""
        # Set up URL's
        m.post(CONST.LOGIN_URL, text=LOGIN.post_response_ok())
        m.get(CONST.OAUTH_TOKEN_URL, text=OAUTH_CLAIMS.get_response_ok())
        m.post(CONST.LOGOUT_URL, text=LOGOUT.post_response_ok())
        m.get(CONST.PANEL_URL, text=PANEL.get_response_ok(mode=CONST.MODE_STANDBY))
        m.get(
            CONST.DEVICES_URL,
            text='['
            + COVER.device(
                devid=COVER.DEVICE_ID,
                status=CONST.STATUS_CLOSED,
                low_battery=False,
                no_response=False,
            )
            + ", "
            + DOORCONTACT.device(
                devid=DOORCONTACT.DEVICE_ID, status=CONST.STATUS_CLOSED
            )
            + ']',
        )

        # Logout to reset everything
        self.abode.logout()

        # Get our devices
        cover = self.abode.get_device(COVER.DEVICE_ID)
        assert cover is not None

        doorcontact = self.abode.get_device(DOORCONTACT.DEVICE_ID)
        assert doorcontact is not None

        # Get the event controller
        events = self.abode.events
        assert events is not None

        callback = Mock()

        # Register our devices
        assert events.add_device_callback([cover, doorcontact], callback)

        # Set up device update URL's
        cover_url = str.replace(CONST.DEVICE_URL, '$DEVID$', COVER.DEVICE_ID)
        m.get(
            cover_url,
            text=COVER.device(
                devid=COVER.DEVICE_ID,
                status=CONST.STATUS_OPEN,
                low_battery=False,
                no_response=False,
            ),
        )

        door_url = str.replace(CONST.DEVICE_URL, '$DEVID$', DOORCONTACT.DEVICE_ID)
        m.get(
            door_url,
            text=DOORCONTACT.device(devid=COVER.DEVICE_ID, status=CONST.STATUS_OPEN),
        )

        # Call our device callback method for our cover
        # pylint: disable=protected-access
        events._on_device_update(cover.device_id)
        callback.assert_called_with(cover)

        # Test that our device updated
        assert cover.status == CONST.STATUS_OPEN

        # Test that our other device didn't update
        assert doorcontact.status == CONST.STATUS_CLOSED

        # Call our device callback method for our door contact
        events._on_device_update(doorcontact.device_id)
        callback.assert_has_calls([call(cover), call(doorcontact)])

        # Test that our door updated now
        assert doorcontact.status == CONST.STATUS_OPEN

    def tests_multi_events_callback(self):
        """Tests that multiple event updates callback correctly."""
        # Get the event controller
        events = self.abode.events
        assert events is not None

        # Create mock callback
        callback = Mock()

        # Register our events
        assert events.add_event_callback(
            [TIMELINE.ALARM_GROUP, TIMELINE.CAPTURE_GROUP], callback
        )

        # Call our events callback method and trigger a capture group event
        # pylint: disable=protected-access
        event_json = json.loads(IRCAMERA.timeline_event())
        events._on_timeline_update(event_json)

        # Ensure our callback was called
        callback.assert_called_with(event_json)

    def tests_multi_timeline_callback(self):
        """Tests that multiple timeline updates callback correctly."""
        # Get the event controller
        events = self.abode.events
        assert events is not None

        # Create mock callback
        callback = Mock()

        # Register our events
        assert events.add_timeline_callback(
            [TIMELINE.CAPTURE_IMAGE, TIMELINE.OPENED], callback
        )

        # Call our events callback method and trigger a capture group event
        # pylint: disable=protected-access
        event_json = json.loads(IRCAMERA.timeline_event())
        events._on_timeline_update(event_json)

        # Ensure our callback was called
        callback.assert_called_with(event_json)

    def tests_automations_callback(self):
        """Tests that automation updates callback correctly."""
        # Get the event controller
        events = self.abode.events
        assert events is not None

        # Create mock callbacks
        automation_callback = Mock()

        # Register our events
        assert events.add_event_callback(
            TIMELINE.AUTOMATION_EDIT_GROUP, automation_callback
        )

        # Call our events callback method and trigger a capture group event
        # pylint: disable=protected-access
        events._on_automation_update('{}')

        # Our capture callback should get one, but our alarm should not
        automation_callback.assert_called_with('{}')



"""Errors for AbodePy."""
USERNAME = (0, "Username must be a non-empty string")

PASSWORD = (1, "Password must be a non-empty string")

REQUEST = (2, "Request failed")

SET_STATUS_DEV_ID = (3, "Device status/level response ID does not match request ID")

SET_STATUS_STATE = (4, "Device status/level value does not match request value")

REFRESH = (5, "Failed to refresh device")

SET_MODE = (6, "Failed to set alarm mode")

SET_MODE_AREA = (7, "Set mode response area does not match request area")

SET_MODE_MODE = (8, "Set mode response mode does not match request mode")

INVALID_ALARM_MODE = (9, "Mode is not of a known alarm mode value")

MISSING_ALARM_MODE = (10, "No alarm mode found in object")

INVALID_DEFAULT_ALARM_MODE = (11, "Default alarm mode must be one of 'home' or 'away'")

INVALID_DEVICE_ID = (12, "The given value is not a device or valid device ID")

INVALID_SETTING = (13, "Setting is not valid")

INVALID_SETTING_VALUE = (14, "Value for setting is not valid")

INVALID_AUTOMATION_REFRESH_RESPONSE = (
    15,
    "Automation refresh response did not match expected values.",
)

INVALID_AUTOMATION_EDIT_RESPONSE = (
    16,
    "Automation edit response did not match expected values.",
)

# DEPRECATED
# TRIGGER_NON_QUICKACTION = (
#     17, "Can not trigger an automation that is not a manual quick-action.")

UNABLE_TO_MAP_DEVICE = (
    18,
    "Unable to map device json to device class - no type tag found.",
)

EVENT_CODE_MISSING = (19, "Event is not valid, start and end event codes are missing.")

EVENT_CODE_MISSING = (20, "Timeline event is not valid, event code missing.")

INVALID_TIMELINE_EVENT = (21, "Timeline event received missing an event code or type.")

EVENT_GROUP_INVALID = (22, "Timeline event group is not valid.")

CAM_IMAGE_REFRESH_NO_FILE = (23, "Camera image refresh did not have a file path.")

CAM_IMAGE_UNEXPECTED_RESPONSE = (24, "Unknown camera image response.")

CAM_IMAGE_NO_LOCATION_HEADER = (
    25,
    "Camera file path did not redirect to image location.",
)

CAM_TIMELINE_EVENT_INVALID = (26, "Timeline event_code invalid - expected 5001.")

CAM_IMAGE_REQUEST_INVALID = (
    27,
    "Received an invalid response from AWS servers for image.",
)

EVENT_DEVICE_INVALID = (
    28,
    "Object given to event registration service is not a device object",
)

SOCKETIO_ERROR = (29, "SocketIO Error Packet Received")

MISSING_CONTROL_URL = (30, "Control URL does not exist in device JSON.")

SET_PRIVACY_MODE = (31, "Device privacy mode value does not match request value.")

MFA_CODE_REQUIRED = (32, "Multifactor authentication code required for login.")

UNKNOWN_MFA_TYPE = (33, "Unknown multifactor authentication type.")



"""AbodePy constants."""

REQUIRED_PYTHON_VER = (3, 6, 0)

PROJECT_GITHUB_USERNAME = 'jaraco'
PROJECT_GITHUB_REPOSITORY = 'jaraco.abode'

CACHE_PATH = './abode.pickle'
COOKIES = "cookies"

ID = 'id'
PASSWORD = 'password'
UUID = 'uuid'
MFA_CODE = 'mfa_code'

# URLS
BASE_URL = 'https://my.goabode.com/'

LOGIN_URL = BASE_URL + 'api/auth2/login'
LOGOUT_URL = BASE_URL + 'api/v1/logout'

OAUTH_TOKEN_URL = BASE_URL + 'api/auth2/claims'

PARAMS_URL = BASE_URL + 'api/v1/devices_beta/'

PANEL_URL = BASE_URL + 'api/v1/panel'

INTEGRATIONS_URL = BASE_URL + 'integrations/v1/devices/'


def get_panel_mode_url(area, mode):
    """Create panel URL."""
    return BASE_URL + 'api/v1/panel/mode/' + area + '/' + mode


DEVICES_URL = BASE_URL + 'api/v1/devices'
DEVICE_URL = BASE_URL + 'api/v1/devices/$DEVID$'

AREAS_URL = BASE_URL + 'api/v1/areas'

SETTINGS_URL = BASE_URL + 'api/v1/panel/setting'
SOUNDS_URL = BASE_URL + 'api/v1/sounds'
SIREN_URL = BASE_URL + 'api/v1/siren'

AUTOMATION_URL = BASE_URL + 'integrations/v1/automations/'
AUTOMATION_ID_URL = AUTOMATION_URL + '$AUTOMATIONID$/'
AUTOMATION_APPLY_URL = AUTOMATION_ID_URL + 'apply'

TIMELINE_IMAGES_ID_URL = (
    BASE_URL
    + 'api/v1/timeline?device_id=$DEVID$&dir=next'
    + '&event_label=Image+Capture&size=1'
)

# NOTIFICATION CONSTANTS
SOCKETIO_URL = 'wss://my.goabode.com/socket.io/'

SOCKETIO_HEADERS = {
    'Origin': 'https://my.goabode.com/',
}

DEVICE_UPDATE_EVENT = 'com.goabode.device.update'
GATEWAY_MODE_EVENT = 'com.goabode.gateway.mode'
TIMELINE_EVENT = 'com.goabode.gateway.timeline'
AUTOMATION_EVENT = 'com.goabode.automation'

# DICTIONARIES
MODE_STANDBY = 'standby'
MODE_HOME = 'home'
MODE_AWAY = 'away'

ALL_MODES = [MODE_STANDBY, MODE_HOME, MODE_AWAY]

ALL_MODES_STR = ", ".join(ALL_MODES)

ARMED = {'home': True, 'away': True, 'standby': False}

STATUS_ONLINE = 'Online'
STATUS_OFFLINE = 'Offline'

STATUS_OPEN = 'Open'
STATUS_OPEN_INT = 1
STATUS_CLOSED = 'Closed'
STATUS_CLOSED_INT = 0

STATUS_LOCKOPEN = 'LockOpen'
STATUS_LOCKOPEN_INT = 0
STATUS_LOCKCLOSED = 'LockClosed'
STATUS_LOCKCLOSED_INT = 1

STATUS_ON = 'On'
STATUS_ON_INT = 1
STATUS_OFF = 'Off'
STATUS_OFF_INT = 0

COLOR_MODE_ON = 0
COLOR_MODE_OFF = 2

# GENERIC ABODE DEVICE TYPES
TYPE_ALARM = "alarm"
TYPE_CAMERA = "camera"
TYPE_CONNECTIVITY = "connectivity"
TYPE_COVER = "cover"
TYPE_LIGHT = "light"
TYPE_LOCK = "lock"
TYPE_MOISTURE = "moisture"
TYPE_MOTION = "motion"
TYPE_OCCUPANCY = "occupancy"
TYPE_OPENING = "door"
TYPE_SENSOR = "sensor"
TYPE_SWITCH = "switch"
TYPE_VALVE = "valve"

TYPE_UNKNOWN_SENSOR = "unknown_sensor"

BINARY_SENSOR_TYPES = [
    TYPE_CONNECTIVITY,
    TYPE_MOISTURE,
    TYPE_MOTION,
    TYPE_OCCUPANCY,
    TYPE_OPENING,
]

# DEVICE TYPE_TAGS
# Alarm
DEVICE_ALARM = 'device_type.alarm'

# Binary Sensors - Connectivity
DEVICE_GLASS_BREAK = 'device_type.glass'
DEVICE_KEYPAD = 'device_type.keypad'
DEVICE_REMOTE_CONTROLLER = 'device_type.remote_controller'
DEVICE_SIREN = 'device_type.siren'
DEVICE_STATUS_DISPLAY = 'device_type.bx'

# Binary Sensors - Opening
DEVICE_DOOR_CONTACT = 'device_type.door_contact'

# Cameras
DEVICE_MOTION_CAMERA = 'device_type.ir_camera'
DEVICE_MOTION_VIDEO_CAMERA = 'device_type.ir_camcoder'
DEVICE_IP_CAM = 'device_type.ipcam'
DEVICE_OUTDOOR_MOTION_CAMERA = 'device_type.out_view'
DEVICE_OUTDOOR_SMART_CAMERA = 'device_type.vdp'

# Covers
DEVICE_SECURE_BARRIER = 'device_type.secure_barrier'

# Dimmers
DEVICE_DIMMER = 'device_type.dimmer'
DEVICE_DIMMER_METER = 'device_type.dimmer_meter'
DEVICE_HUE = 'device_type.hue'

# Locks
DEVICE_DOOR_LOCK = 'device_type.door_lock'

# Moisture
DEVICE_WATER_SENSOR = 'device_type.water_sensor'

# Switches
DEVICE_SWITCH = 'device_type.switch'
DEVICE_NIGHT_SWITCH = 'device_type.night_switch'
DEVICE_POWER_SWITCH_SENSOR = 'device_type.power_switch_sensor'
DEVICE_POWER_SWITCH_METER = 'device_type.power_switch_meter'

# Water Valve
DEVICE_VALVE = 'device_type.valve'

# Unknown Sensor
# These device types are all considered 'occupancy' but could apparently
# also be multi-sensors based on their json.
DEVICE_ROOM_SENSOR = 'device_type.room_sensor'
DEVICE_TEMPERATURE_SENSOR = 'device_type.temperature_sensor'
DEVICE_MULTI_SENSOR = 'device_type.lm'  # LM = LIGHT MOTION?
DEVICE_PIR = 'device_type.pir'  # Passive Infrared Occupancy?
DEVICE_POVS = 'device_type.povs'

STATUSES_KEY = 'statuses'
TEMP_STATUS_KEY = 'temperature'
LUX_STATUS_KEY = 'lux'
HUMI_STATUS_KEY = 'humidity'
SENSOR_KEYS = [TEMP_STATUS_KEY, LUX_STATUS_KEY, HUMI_STATUS_KEY]

UNIT_CELSIUS = '°C'
UNIT_FAHRENHEIT = '°F'
UNIT_PERCENT = '%'
UNIT_LUX = 'lx'
LUX = 'lux'

BRIGHTNESS_KEY = 'statusEx'


def get_generic_type(type_tag):
    """Map type tag to generic type."""
    return {
        # Alarm
        DEVICE_ALARM: TYPE_ALARM,
        # Binary Sensors - Connectivity
        DEVICE_GLASS_BREAK: TYPE_CONNECTIVITY,
        DEVICE_KEYPAD: TYPE_CONNECTIVITY,
        DEVICE_REMOTE_CONTROLLER: TYPE_CONNECTIVITY,
        DEVICE_SIREN: TYPE_CONNECTIVITY,
        DEVICE_STATUS_DISPLAY: TYPE_CONNECTIVITY,
        # Binary Sensors - Opening
        DEVICE_DOOR_CONTACT: TYPE_OPENING,
        # Cameras
        DEVICE_MOTION_CAMERA: TYPE_CAMERA,
        DEVICE_MOTION_VIDEO_CAMERA: TYPE_CAMERA,
        DEVICE_IP_CAM: TYPE_CAMERA,
        DEVICE_OUTDOOR_MOTION_CAMERA: TYPE_CAMERA,
        DEVICE_OUTDOOR_SMART_CAMERA: TYPE_CAMERA,
        # Covers
        DEVICE_SECURE_BARRIER: TYPE_COVER,
        # Lights (Dimmers)
        DEVICE_DIMMER: TYPE_LIGHT,
        DEVICE_DIMMER_METER: TYPE_LIGHT,
        DEVICE_HUE: TYPE_LIGHT,
        # Locks
        DEVICE_DOOR_LOCK: TYPE_LOCK,
        # Moisture
        DEVICE_WATER_SENSOR: TYPE_CONNECTIVITY,
        # Switches
        DEVICE_SWITCH: TYPE_SWITCH,
        DEVICE_NIGHT_SWITCH: TYPE_SWITCH,
        DEVICE_POWER_SWITCH_SENSOR: TYPE_SWITCH,
        DEVICE_POWER_SWITCH_METER: TYPE_SWITCH,
        # Water Valve
        DEVICE_VALVE: TYPE_VALVE,
        # Unknown Sensors
        # More data needed to determine type
        DEVICE_ROOM_SENSOR: TYPE_UNKNOWN_SENSOR,
        DEVICE_TEMPERATURE_SENSOR: TYPE_UNKNOWN_SENSOR,
        DEVICE_MULTI_SENSOR: TYPE_UNKNOWN_SENSOR,
        DEVICE_PIR: TYPE_UNKNOWN_SENSOR,
        DEVICE_POVS: TYPE_UNKNOWN_SENSOR,
    }.get(type_tag.lower(), None)


# Constants to be used to fill our imaginary alarm device
ALARM_NAME = "Abode Alarm"
ALARM_DEVICE_ID = "area_"
ALARM_TYPE = "Alarm"

# SETTINGS
SETTING_CAMERA_RESOLUTION = 'ircamera_resolution_t'
SETTING_CAMERA_GRAYSCALE = 'ircamera_gray_t'
SETTING_SILENCE_SOUNDS = 'beeper_mute'

PANEL_SETTINGS = [
    SETTING_CAMERA_RESOLUTION,
    SETTING_CAMERA_GRAYSCALE,
    SETTING_SILENCE_SOUNDS,
]

SETTING_ENTRY_DELAY_AWAY = 'away_entry_delay'
SETTING_EXIT_DELAY_AWAY = 'away_exit_delay'
SETTING_ENTRY_DELAY_HOME = 'home_entry_delay'
SETTING_EXIT_DELAY_HOME = 'home_exit_delay'

AREA_SETTINGS = [
    SETTING_ENTRY_DELAY_AWAY,
    SETTING_EXIT_DELAY_AWAY,
    SETTING_ENTRY_DELAY_HOME,
    SETTING_EXIT_DELAY_HOME,
]

SETTING_DOOR_CHIME = 'door_chime'
SETTING_WARNING_BEEP = 'warning_beep'
SETTING_ENTRY_BEEP_AWAY = 'entry_beep_away'
SETTING_EXIT_BEEP_AWAY = 'exit_beep_away'
SETTING_ENTRY_BEEP_HOME = 'entry_beep_home'
SETTING_EXIT_BEEP_HOME = 'exit_beep_home'
SETTING_CONFIRM_SOUND = 'confirm_snd'
SETTING_ALARM_LENGTH = 'alarm_len'
SETTING_FINAL_BEEPS = 'final_beep'

SOUND_SETTINGS = [
    SETTING_DOOR_CHIME,
    SETTING_WARNING_BEEP,
    SETTING_ENTRY_BEEP_AWAY,
    SETTING_EXIT_BEEP_AWAY,
    SETTING_ENTRY_BEEP_HOME,
    SETTING_EXIT_BEEP_HOME,
    SETTING_CONFIRM_SOUND,
    SETTING_ALARM_LENGTH,
    SETTING_FINAL_BEEPS,
]

SETTING_SIREN_ENTRY_EXIT_SOUNDS = "entry"
SETTING_SIREN_TAMPER_SOUNDS = "tamper"
SETTING_SIREN_CONFIRM_SOUNDS = "confirm"

SIREN_SETTINGS = [
    SETTING_SIREN_ENTRY_EXIT_SOUNDS,
    SETTING_SIREN_TAMPER_SOUNDS,
    SETTING_SIREN_CONFIRM_SOUNDS,
]

ALL_SETTINGS = PANEL_SETTINGS + AREA_SETTINGS + SOUND_SETTINGS + SIREN_SETTINGS


# SETTING VALUES
SETTING_CAMERA_RES_320_240 = '0'
SETTING_CAMERA_RES_640_480 = '2'

SETTING_ALL_CAMERA_RES = [SETTING_CAMERA_RES_320_240, SETTING_CAMERA_RES_640_480]

SETTING_DISABLE = '0'
SETTING_ENABLE = '1'

SETTING_DISABLE_ENABLE = [SETTING_DISABLE, SETTING_ENABLE]

SETTING_ENTRY_EXIT_DELAY_DISABLE = '0'
SETTING_ENTRY_EXIT_DELAY_10SEC = '10'
SETTING_ENTRY_EXIT_DELAY_20SEC = '20'
SETTING_ENTRY_EXIT_DELAY_30SEC = '30'
SETTING_ENTRY_EXIT_DELAY_1MIN = '60'
SETTING_ENTRY_EXIT_DELAY_2MIN = '120'
SETTING_ENTRY_EXIT_DELAY_3MIN = '180'
SETTING_ENTRY_EXIT_DELAY_4MIN = '240'

ALL_SETTING_ENTRY_EXIT_DELAY = [
    SETTING_ENTRY_EXIT_DELAY_DISABLE,
    SETTING_ENTRY_EXIT_DELAY_10SEC,
    SETTING_ENTRY_EXIT_DELAY_20SEC,
    SETTING_ENTRY_EXIT_DELAY_30SEC,
    SETTING_ENTRY_EXIT_DELAY_1MIN,
    SETTING_ENTRY_EXIT_DELAY_2MIN,
    SETTING_ENTRY_EXIT_DELAY_3MIN,
    SETTING_ENTRY_EXIT_DELAY_4MIN,
]

VALID_SETTING_EXIT_AWAY = [
    SETTING_ENTRY_EXIT_DELAY_30SEC,
    SETTING_ENTRY_EXIT_DELAY_1MIN,
    SETTING_ENTRY_EXIT_DELAY_2MIN,
    SETTING_ENTRY_EXIT_DELAY_3MIN,
    SETTING_ENTRY_EXIT_DELAY_4MIN,
]

SETTING_SOUND_OFF = 'none'
SETTING_SOUND_LOW = 'normal'
SETTING_SOUND_HIGH = 'loud'

ALL_SETTING_SOUND = [SETTING_SOUND_OFF, SETTING_SOUND_LOW, SETTING_SOUND_HIGH]

VALID_SOUND_SETTINGS = [
    SETTING_DOOR_CHIME,
    SETTING_WARNING_BEEP,
    SETTING_ENTRY_BEEP_AWAY,
    SETTING_EXIT_BEEP_AWAY,
    SETTING_ENTRY_BEEP_HOME,
    SETTING_EXIT_BEEP_HOME,
    SETTING_CONFIRM_SOUND,
]

SETTING_ALARM_LENGTH_DISABLE = '0'
SETTING_ALARM_LENGTH_1MIN = '60'
SETTING_ALARM_LENGTH_2MIN = '120'
SETTING_ALARM_LENGTH_3MIN = '180'
SETTING_ALARM_LENGTH_4MIN = '240'
SETTING_ALARM_LENGTH_5MIN = '300'
SETTING_ALARM_LENGTH_6MIN = '360'
SETTING_ALARM_LENGTH_7MIN = '420'
SETTING_ALARM_LENGTH_8MIN = '480'
SETTING_ALARM_LENGTH_9MIN = '540'
SETTING_ALARM_LENGTH_10MIN = '600'
SETTING_ALARM_LENGTH_11MIN = '660'
SETTING_ALARM_LENGTH_12MIN = '720'
SETTING_ALARM_LENGTH_13MIN = '780'
SETTING_ALARM_LENGTH_14MIN = '840'
SETTING_ALARM_LENGTH_15MIN = '900'

ALL_SETTING_ALARM_LENGTH = [
    SETTING_ALARM_LENGTH_DISABLE,
    SETTING_ALARM_LENGTH_1MIN,
    SETTING_ALARM_LENGTH_2MIN,
    SETTING_ALARM_LENGTH_3MIN,
    SETTING_ALARM_LENGTH_4MIN,
    SETTING_ALARM_LENGTH_5MIN,
    SETTING_ALARM_LENGTH_6MIN,
    SETTING_ALARM_LENGTH_7MIN,
    SETTING_ALARM_LENGTH_8MIN,
    SETTING_ALARM_LENGTH_9MIN,
    SETTING_ALARM_LENGTH_10MIN,
    SETTING_ALARM_LENGTH_11MIN,
    SETTING_ALARM_LENGTH_12MIN,
    SETTING_ALARM_LENGTH_13MIN,
    SETTING_ALARM_LENGTH_14MIN,
    SETTING_ALARM_LENGTH_15MIN,
]

SETTING_FINAL_BEEPS_DISABLE = '0'
SETTING_FINAL_BEEPS_3SEC = '3'
SETTING_FINAL_BEEPS_4SEC = '4'
SETTING_FINAL_BEEPS_5SEC = '5'
SETTING_FINAL_BEEPS_6SEC = '6'
SETTING_FINAL_BEEPS_7SEC = '7'
SETTING_FINAL_BEEPS_8SEC = '8'
SETTING_FINAL_BEEPS_9SEC = '9'
SETTING_FINAL_BEEPS_10SEC = '10'

ALL_SETTING_FINAL_BEEPS = [
    SETTING_FINAL_BEEPS_DISABLE,
    SETTING_FINAL_BEEPS_3SEC,
    SETTING_FINAL_BEEPS_4SEC,
    SETTING_FINAL_BEEPS_5SEC,
    SETTING_FINAL_BEEPS_6SEC,
    SETTING_FINAL_BEEPS_7SEC,
    SETTING_FINAL_BEEPS_8SEC,
    SETTING_FINAL_BEEPS_9SEC,
    SETTING_FINAL_BEEPS_10SEC,
]



"""Mock Abode Panel Response."""

import jaraco.abode.helpers.constants as CONST


def get_response_ok(
    mode=CONST.MODE_STANDBY, battery=False, is_cellular=False, mac='00:11:22:33:44:55'
):
    """Return panel response json."""
    return (
        '''{
       "version":"ABGW 0.0.2.17F ABGW-L1-XA36J 3.1.2.6.1 Z-Wave 3.95",
       "report_account":"5555",
       "online":"1",
       "initialized":"1",
       "net_version":"ABGW 0.0.2.17F",
       "rf_version":"ABGW-L1-XA36J",
       "zigbee_version":"3.1.2.6.1",
       "z_wave_version":"Z-Wave 3.95",
       "timezone":"America/New_York",
       "ac_fail":"0",
       "battery":"'''
        + str(int(battery))
        + '''",
       "ip":"192.168.1.1",
       "jam":"0",
       "rssi":"2",
       "setup_zone_1":"1",
       "setup_zone_2":"1",
       "setup_zone_3":"1",
       "setup_zone_4":"1",
       "setup_zone_5":"1",
       "setup_zone_6":"1",
       "setup_zone_7":"1",
       "setup_zone_8":"1",
       "setup_zone_9":"1",
       "setup_zone_10":"1",
       "setup_gateway":"1",
       "setup_contacts":"1",
       "setup_billing":"1",
       "setup_users":"1",
       "is_cellular":"'''
        + str(int(is_cellular))
        + '''",
       "plan_set_id":"1",
       "dealer_id":"0",
       "tz_diff":"-04:00",
       "is_demo":"0",
       "rf51_version":"ABGW-L1-XA36J",
       "model":"L1",
       "mac":"'''
        + mac
        + '''",
       "xml_version":"3",
       "dealer_name":"abode",
       "id":"0",
       "dealer_address":"2625 Middlefield Road #900 Palo Alto CA 94306",
       "dealer_domain":"https://my.goabode.com",
       "domain_alias":"https://test.goabode.com",
       "dealer_support_url":"https://support.goabode.com",
       "app_launch_url":"https://goabode.app.link/abode",
       "has_wifi":"0",
       "mode":{
          "area_1":"'''
        + mode
        + '''",
          "area_2":"standby"
       }
    }'''
    )


def put_response_ok(area='1', mode=CONST.MODE_STANDBY):
    """Return panel mode response json."""
    return '{"area": "' + area + '", "mode": "' + mode + '"}'



"""
Mock Internal Alarm Device.

AbodePy creates an internal device that is a standard panel response that
includes a few required device fields. This is so that we can easily translate
the panel/alarm itself into a Home Assistant device.
"""
import json
import jaraco.abode.helpers.constants as CONST

import tests.mock.panel as PANEL


def device(area='1', panel=PANEL.get_response_ok(mode=CONST.MODE_STANDBY)):
    """Alarm mock device."""
    alarm = json.loads(panel)
    alarm['name'] = CONST.ALARM_NAME
    alarm['id'] = CONST.ALARM_DEVICE_ID + area
    alarm['type'] = CONST.ALARM_TYPE
    alarm['type_tag'] = CONST.DEVICE_ALARM
    alarm['generic_type'] = CONST.TYPE_ALARM
    alarm['uuid'] = alarm.get('mac').replace(':', '').lower()

    return alarm



"""Test the Abode device classes."""
import unittest

import requests_mock

import jaraco.abode
import jaraco.abode.helpers.constants as CONST

import tests.mock.login as LOGIN
import tests.mock.oauth_claims as OAUTH_CLAIMS
import tests.mock.logout as LOGOUT
import tests.mock.panel as PANEL
import tests.mock.devices as DEVICES
import tests.mock.devices.hue as HUE
import pytest


USERNAME = 'foobar'
PASSWORD = 'deadbeef'


class TestHue(unittest.TestCase):
    """Test the AbodePy light device with Hue."""

    def setUp(self):
        """Set up Abode module."""
        self.abode = jaraco.abode.Abode(
            username=USERNAME, password=PASSWORD, disable_cache=True
        )

    def tearDown(self):
        """Clean up after test."""
        self.abode = None

    @requests_mock.mock()
    def tests_hue_device_properties(self, m):
        """Tests that hue light devices properties work as expected."""
        # Set up URL's
        m.post(CONST.LOGIN_URL, text=LOGIN.post_response_ok())
        m.get(CONST.OAUTH_TOKEN_URL, text=OAUTH_CLAIMS.get_response_ok())
        m.post(CONST.LOGOUT_URL, text=LOGOUT.post_response_ok())
        m.get(CONST.PANEL_URL, text=PANEL.get_response_ok(mode=CONST.MODE_STANDBY))
        m.get(
            CONST.DEVICES_URL,
            text=HUE.device(
                devid=HUE.DEVICE_ID,
                status=CONST.STATUS_OFF,
                level=0,
                saturation=57,
                hue=60,
                color_temp=6536,
                color_mode=CONST.COLOR_MODE_ON,
                low_battery=False,
                no_response=False,
            ),
        )

        # Logout to reset everything
        self.abode.logout()

        # Get our dimmer
        device = self.abode.get_device(HUE.DEVICE_ID)

        # Test our device
        assert device is not None
        assert device.status == CONST.STATUS_OFF
        assert device.brightness == "0"
        assert device.color == (60, 57)  # (hue, saturation)
        assert device.color_temp == 6536
        assert device.has_brightness
        assert device.is_dimmable
        assert device.has_color
        assert device.is_color_capable
        assert not device.battery_low
        assert not device.no_response
        assert not device.is_on

        # Set up our direct device get url
        device_url = str.replace(CONST.DEVICE_URL, '$DEVID$', HUE.DEVICE_ID)

        # Change device properties
        m.get(
            device_url,
            text=HUE.device(
                devid=HUE.DEVICE_ID,
                status=CONST.STATUS_ON,
                level=45,
                saturation=22,
                hue=104,
                color_temp=4000,
                color_mode=CONST.COLOR_MODE_OFF,
                low_battery=True,
                no_response=True,
            ),
        )

        # Refesh device and test changes
        device.refresh()

        assert device.status == CONST.STATUS_ON
        assert device.color == (104, 22)  # (hue, saturation)
        assert device.color_temp == 4000
        assert device.has_brightness
        assert device.is_dimmable
        assert not device.has_color
        assert device.is_color_capable
        assert device.battery_low
        assert device.no_response
        assert device.is_on

    @requests_mock.mock()
    def tests_hue_status_changes(self, m):
        """Tests that hue device changes work as expected."""
        # Set up URL's
        m.post(CONST.LOGIN_URL, text=LOGIN.post_response_ok())
        m.get(CONST.OAUTH_TOKEN_URL, text=OAUTH_CLAIMS.get_response_ok())
        m.post(CONST.LOGOUT_URL, text=LOGOUT.post_response_ok())
        m.get(CONST.PANEL_URL, text=PANEL.get_response_ok(mode=CONST.MODE_STANDBY))
        m.get(
            CONST.DEVICES_URL,
            text=HUE.device(
                devid=HUE.DEVICE_ID,
                status=CONST.STATUS_OFF,
                level=0,
                saturation=57,
                hue=60,
                color_temp=6536,
                color_mode=CONST.COLOR_MODE_ON,
                low_battery=False,
                no_response=False,
            ),
        )

        # Logout to reset everything
        self.abode.logout()

        # Get our hue device
        device = self.abode.get_device(HUE.DEVICE_ID)

        # Test that we have our device
        assert device is not None
        assert device.status == CONST.STATUS_OFF
        assert not device.is_on

        # Set up control url response
        control_url = CONST.BASE_URL + HUE.CONTROL_URL
        m.put(
            control_url,
            text=DEVICES.status_put_response_ok(
                devid=HUE.DEVICE_ID, status=CONST.STATUS_ON_INT
            ),
        )

        # Change the mode to "on"
        assert device.switch_on()
        assert device.status == CONST.STATUS_ON
        assert device.is_on

        # Change response
        m.put(
            control_url,
            text=DEVICES.status_put_response_ok(
                devid=HUE.DEVICE_ID, status=CONST.STATUS_OFF_INT
            ),
        )

        # Change the mode to "off"
        assert device.switch_off()
        assert device.status == CONST.STATUS_OFF
        assert not device.is_on

        # Test that an invalid status response throws exception
        m.put(
            control_url,
            text=DEVICES.status_put_response_ok(
                devid=HUE.DEVICE_ID, status=CONST.STATUS_OFF_INT
            ),
        )

        with pytest.raises(jaraco.abode.AbodeException):
            device.switch_on()

    @requests_mock.mock()
    def tests_hue_color_temp_changes(self, m):
        """Tests that hue device color temp changes work as expected."""
        # Set up URL's
        m.post(CONST.LOGIN_URL, text=LOGIN.post_response_ok())
        m.get(CONST.OAUTH_TOKEN_URL, text=OAUTH_CLAIMS.get_response_ok())
        m.post(CONST.LOGOUT_URL, text=LOGOUT.post_response_ok())
        m.get(CONST.PANEL_URL, text=PANEL.get_response_ok(mode=CONST.MODE_STANDBY))
        m.get(
            CONST.DEVICES_URL,
            text=HUE.device(
                devid=HUE.DEVICE_ID,
                status=CONST.STATUS_OFF,
                level=0,
                saturation=57,
                hue=60,
                color_temp=6536,
                color_mode=CONST.COLOR_MODE_ON,
                low_battery=False,
                no_response=False,
            ),
        )

        # Logout to reset everything
        self.abode.logout()

        # Get our hue device
        device = self.abode.get_device(HUE.DEVICE_ID)

        # Test that we have our device
        assert device is not None
        assert device.status == CONST.STATUS_OFF
        assert not device.is_on
        assert device.color_temp == 6536

        # Set up integrations url response
        m.post(
            HUE.INTEGRATIONS_URL,
            text=HUE.color_temp_post_response_ok(devid=HUE.DEVICE_ID, color_temp=5554),
        )

        # Change the color temp
        assert device.set_color_temp(5554)
        assert device.color_temp == 5554

        # Change response
        m.post(
            HUE.INTEGRATIONS_URL,
            text=HUE.color_temp_post_response_ok(devid=HUE.DEVICE_ID, color_temp=4434),
        )

        # Change the color to something that mismatches
        assert device.set_color_temp(4436)

        # Assert that the color is set to the response color
        assert device.color_temp == 4434

        # Test that an invalid ID in response throws exception
        m.post(
            HUE.INTEGRATIONS_URL,
            text=HUE.color_temp_post_response_ok(
                devid=(HUE.DEVICE_ID + "23"), color_temp=4434
            ),
        )

        with pytest.raises(jaraco.abode.AbodeException):
            device.set_color_temp(4434)

    @requests_mock.mock()
    def tests_hue_color_changes(self, m):
        """Tests that hue device color changes work as expected."""
        # Set up URL's
        m.post(CONST.LOGIN_URL, text=LOGIN.post_response_ok())
        m.get(CONST.OAUTH_TOKEN_URL, text=OAUTH_CLAIMS.get_response_ok())
        m.post(CONST.LOGOUT_URL, text=LOGOUT.post_response_ok())
        m.get(CONST.PANEL_URL, text=PANEL.get_response_ok(mode=CONST.MODE_STANDBY))
        m.get(
            CONST.DEVICES_URL,
            text=HUE.device(
                devid=HUE.DEVICE_ID,
                status=CONST.STATUS_OFF,
                level=0,
                saturation=57,
                hue=60,
                color_temp=6536,
                color_mode=CONST.COLOR_MODE_ON,
                low_battery=False,
                no_response=False,
            ),
        )

        # Logout to reset everything
        self.abode.logout()

        # Get our hue device
        device = self.abode.get_device(HUE.DEVICE_ID)

        # Test that we have our device
        assert device is not None
        assert device.status == CONST.STATUS_OFF
        assert not device.is_on
        assert device.color == (60, 57)  # (hue, saturation)

        # Set up integrations url response
        m.post(
            HUE.INTEGRATIONS_URL,
            text=HUE.color_post_response_ok(devid=HUE.DEVICE_ID, hue=70, saturation=80),
        )

        # Change the color temp
        assert device.set_color((70, 80))
        assert device.color == (70, 80)  # (hue, saturation)

        # Change response
        m.post(
            HUE.INTEGRATIONS_URL,
            text=HUE.color_post_response_ok(devid=HUE.DEVICE_ID, hue=55, saturation=85),
        )

        # Change the color to something that mismatches
        assert device.set_color((44, 44))

        # Assert that the color is set to the response color
        assert device.color == (55, 85)  # (hue, saturation)

        # Test that an invalid ID in response throws exception
        m.post(
            HUE.INTEGRATIONS_URL,
            text=HUE.color_post_response_ok(
                devid=(HUE.DEVICE_ID + "23"), hue=55, saturation=85
            ),
        )

        with pytest.raises(jaraco.abode.AbodeException):
            device.set_color((44, 44))



"""Mock Abode Logout Response."""


def post_response_ok():
    """Return the successful logout response json."""
    return '{"code":200,"message":"Logout successful."}'


def post_response_bad_request():
    """Return the failed logout response json."""
    return '''
        {
            "code":400,"message":"Some logout error occurred.",
            "detail":null
        }'''



"""Mock Abode Power Switch Meter Device."""
import jaraco.abode.helpers.constants as CONST

DEVICE_ID = 'ZW:00000077'
CONTROL_URL = 'api/v1/control/power_switch/' + DEVICE_ID


def device(
    devid=DEVICE_ID, status=CONST.STATUS_OFF, low_battery=False, no_response=False
):
    """Power switch mock device."""
    return (
        '''
        {
           "id":"'''
        + devid
        + '''",
           "type_tag":"device_type.power_switch_meter",
           "type":"Power Switch Meter",
           "name":"Aeon Switch",
           "area":"1",
           "zone":"32",
           "sort_order":null,
           "is_window":"",
           "bypass":"0",
           "schar_24hr":"0",
           "sresp_mode_0":"0",
           "sresp_entry_0":"0",
           "sresp_exit_0":"0",
           "sresp_mode_1":"0",
           "sresp_entry_1":"0",
           "sresp_exit_1":"0",
           "sresp_mode_2":"0",
           "sresp_entry_2":"0",
           "sresp_exit_2":"0",
           "sresp_mode_3":"0",
           "sresp_entry_3":"0",
           "sresp_exit_3":"0",
           "capture_mode":null,
           "origin":"abode",
           "control_url":"'''
        + CONTROL_URL
        + '''",
           "deep_link":null,
           "status_color":"#5cb85c",
           "faults":{
              "low_battery":'''
        + str(int(low_battery))
        + ''',
              "tempered":0,
              "supervision":0,
              "out_of_order":0,
              "no_response":'''
        + str(int(no_response))
        + '''
           },
           "status":"'''
        + status
        + '''",
           "statuses":{
              "hvac_mode":null
           },
           "status_ex":"",
           "actions":[
              {
                 "label":"Switch off",
                 "value":"a=1&z=32&sw=off;"
              },
              {
                 "label":"Switch on",
                 "value":"a=1&z=32&sw=on;"
              },
              {
                 "label":"Toggle",
                 "value":"a=1&z=32&sw=toggle;"
              },
              {
                 "label":"Switch on for 5 min",
                 "value":"a=1&z=32&sw=on&pd=5;"
              },
              {
                 "label":"Switch on for 10 min",
                 "value":"a=1&z=32&sw=on&pd=10;"
              },
              {
                 "label":"Switch on for 15 min",
                 "value":"a=1&z=32&sw=on&pd=15;"
              },
              {
                 "label":"Switch on for 20 min",
                 "value":"a=1&z=32&sw=on&pd=20;"
              },
              {
                 "label":"Switch on for 25 min",
                 "value":"a=1&z=32&sw=on&pd=25;"
              },
              {
                 "label":"Switch on for 30 min",
                 "value":"a=1&z=32&sw=on&pd=30;"
              },
              {
                 "label":"Switch on for 45 min",
                 "value":"a=1&z=32&sw=on&pd=45;"
              },
              {
                 "label":"Switch on for 1 hour",
                 "value":"a=1&z=32&sw=on&pd=60;"
              },
              {
                 "label":"Switch on for 2 hours",
                 "value":"a=1&z=32&sw=on&pd=120;"
              },
              {
                 "label":"Switch on for 5 hours",
                 "value":"a=1&z=32&sw=on&pd=300;"
              },
              {
                 "label":"Switch on for 8 hours",
                 "value":"a=1&z=32&sw=on&pd=480;"
              }
           ],
           "status_icons":[
           ],
           "icon":"assets/icons/plug.svg"
        }'''
    )



"""Abode alarm device."""
import json
import logging

from ..exceptions import AbodeException

from ..devices.switch import AbodeDevice, AbodeSwitch
from ..helpers import constants as CONST
from ..helpers import errors as ERROR

_LOGGER = logging.getLogger(__name__)


def create_alarm(panel_json, abode, area='1'):
    """Create a new alarm device from a panel response."""
    panel_json['name'] = CONST.ALARM_NAME
    panel_json['id'] = CONST.ALARM_DEVICE_ID + area
    panel_json['type'] = CONST.ALARM_TYPE
    panel_json['type_tag'] = CONST.DEVICE_ALARM
    panel_json['generic_type'] = CONST.TYPE_ALARM
    panel_json['uuid'] = panel_json.get('mac').replace(':', '').lower()

    return AbodeAlarm(panel_json, abode, area)


class AbodeAlarm(AbodeSwitch):
    """Class to represent the Abode alarm as a device."""

    def __init__(self, json_obj, abode, area='1'):
        """Set up Abode alarm device."""
        AbodeSwitch.__init__(self, json_obj, abode)
        self._area = area

    def set_mode(self, mode):
        """Set Abode alarm mode."""
        if not mode:
            raise AbodeException(ERROR.MISSING_ALARM_MODE)

        if mode.lower() not in CONST.ALL_MODES:
            raise AbodeException(ERROR.INVALID_ALARM_MODE, CONST.ALL_MODES)

        mode = mode.lower()

        response = self._abode.send_request(
            "put", CONST.get_panel_mode_url(self._area, mode)
        )

        _LOGGER.debug("Set Alarm Home Response: %s", response.text)

        response_object = json.loads(response.text)

        if response_object['area'] != self._area:
            raise AbodeException(ERROR.SET_MODE_AREA)

        if response_object['mode'] != mode:
            raise AbodeException(ERROR.SET_MODE_MODE)

        self._json_state['mode'][(self.device_id)] = response_object['mode']

        _LOGGER.info(
            "Set alarm %s mode to: %s", self._device_id, response_object['mode']
        )

        return True

    def set_home(self):
        """Arm Abode to home mode."""
        return self.set_mode(CONST.MODE_HOME)

    def set_away(self):
        """Arm Abode to home mode."""
        return self.set_mode(CONST.MODE_AWAY)

    def set_standby(self):
        """Arm Abode to stay mode."""
        return self.set_mode(CONST.MODE_STANDBY)

    def switch_on(self):
        """Arm Abode to default mode."""
        return self.set_mode(self._abode.default_mode)

    def switch_off(self):
        """Arm Abode to home mode."""
        return self.set_standby()

    def refresh(self, url=CONST.PANEL_URL):
        """Refresh the alarm device."""
        response_object = AbodeDevice.refresh(self, url)
        # pylint: disable=W0212
        self._abode._panel.update(response_object[0])

        return response_object

    @property
    def is_on(self):
        """Is alarm armed."""
        return self.mode in (CONST.MODE_HOME, CONST.MODE_AWAY)

    @property
    def is_standby(self):
        """Is alarm in standby mode."""
        return self.mode == CONST.MODE_STANDBY

    @property
    def is_home(self):
        """Is alarm in home mode."""
        return self.mode == CONST.MODE_HOME

    @property
    def is_away(self):
        """Is alarm in away mode."""
        return self.mode == CONST.MODE_AWAY

    @property
    def mode(self):
        """Get alarm mode."""
        mode = self.get_value('mode').get(self.device_id, None)

        return mode.lower()

    @property
    def status(self):
        """To match existing property."""
        return self.mode

    @property
    def battery(self):
        """Return true if base station on battery backup."""
        return int(self._json_state.get('battery', '0')) == 1

    @property
    def is_cellular(self):
        """Return true if base station on cellular backup."""
        return int(self._json_state.get('is_cellular', '0')) == 1

    @property
    def mac_address(self):
        """Get the hub mac address."""
        return self._json_state.get('mac')



"""The exceptions used by AbodePy."""


class AbodeException(Exception):
    """Class to throw general abode exception."""

    def __init__(self, error, details=None):
        """Initialize AbodeException."""
        # Call the base class constructor with the parameters it needs
        super(AbodeException, self).__init__(error[1])

        self.errcode = error[0]
        self.message = error[1]
        self.details = details


class AbodeAuthenticationException(AbodeException):
    """Class to throw authentication exception."""


class SocketIOException(AbodeException):
    """Class to throw SocketIO Error exception."""



"""Abode binary sensor device."""

from ..devices import AbodeDevice
from ..helpers import constants as CONST


class AbodeBinarySensor(AbodeDevice):
    """Class to represent an on / off, online/offline sensor."""

    @property
    def is_on(self):
        """
        Get sensor state.

        Assume offline or open (worst case).
        """
        if self._type == 'Occupancy':
            return self.status not in CONST.STATUS_ONLINE
        return self.status not in (
            CONST.STATUS_OFF,
            CONST.STATUS_OFFLINE,
            CONST.STATUS_CLOSED,
        )



"""Abode lock device."""

from ..devices import AbodeDevice
from ..helpers import constants as CONST


class AbodeLock(AbodeDevice):
    """Class to represent a door lock."""

    def lock(self):
        """Lock the device."""
        success = self.set_status(CONST.STATUS_LOCKCLOSED_INT)

        if success:
            self._json_state['status'] = CONST.STATUS_LOCKCLOSED

        return success

    def unlock(self):
        """Unlock the device."""
        success = self.set_status(CONST.STATUS_LOCKOPEN_INT)

        if success:
            self._json_state['status'] = CONST.STATUS_LOCKOPEN

        return success

    @property
    def is_locked(self):
        """
        Get locked state.

        Err on side of caution, assume if lock isn't closed then it's open.
        """
        return self.status in CONST.STATUS_LOCKCLOSED



"""Test the Abode device classes."""
import unittest

import requests_mock

import jaraco.abode
import jaraco.abode.helpers.constants as CONST

import tests.mock.login as LOGIN
import tests.mock.oauth_claims as OAUTH_CLAIMS
import tests.mock.logout as LOGOUT
import tests.mock.panel as PANEL
import tests.mock.devices.lm as LM


USERNAME = 'foobar'
PASSWORD = 'deadbeef'


class TestLM(unittest.TestCase):
    """Test the AbodePy sensor class/LM."""

    def setUp(self):
        """Set up Abode module."""
        self.abode = jaraco.abode.Abode(
            username=USERNAME, password=PASSWORD, disable_cache=True
        )

    def tearDown(self):
        """Clean up after test."""
        self.abode = None

    @requests_mock.mock()
    def tests_cover_lm_properties(self, m):
        """Tests that sensor/LM devices properties work as expected."""
        # Set up URL's
        m.post(CONST.LOGIN_URL, text=LOGIN.post_response_ok())
        m.get(CONST.OAUTH_TOKEN_URL, text=OAUTH_CLAIMS.get_response_ok())
        m.post(CONST.LOGOUT_URL, text=LOGOUT.post_response_ok())
        m.get(CONST.PANEL_URL, text=PANEL.get_response_ok(mode=CONST.MODE_STANDBY))
        m.get(
            CONST.DEVICES_URL,
            text=LM.device(
                devid=LM.DEVICE_ID,
                status='72 °F',
                temp='72 °F',
                lux='14 lx',
                humidity='34 %',
                low_battery=False,
                no_response=False,
            ),
        )

        # Logout to reset everything
        self.abode.logout()

        # Get our power switch
        device = self.abode.get_device(LM.DEVICE_ID)

        # Test our device
        assert device is not None
        assert device.status == '72 °F'
        assert not device.battery_low
        assert not device.no_response
        assert device.has_temp
        assert device.has_humidity
        assert device.has_lux
        assert device.temp == 72
        assert device.temp_unit == '°F'
        assert device.humidity == 34
        assert device.humidity_unit == '%'
        assert device.lux == 14
        assert device.lux_unit == 'lux'

        # Set up our direct device get url
        device_url = str.replace(CONST.DEVICE_URL, '$DEVID$', LM.DEVICE_ID)

        # Change device properties
        m.get(
            device_url,
            text=LM.device(
                devid=LM.DEVICE_ID,
                status='12 °C',
                temp='12 °C',
                lux='100 lx',
                humidity='100 %',
                low_battery=True,
                no_response=True,
            ),
        )

        # Refesh device and test changes
        device.refresh()

        assert device.status == '12 °C'
        assert device.battery_low
        assert device.no_response
        assert device.has_temp
        assert device.has_humidity
        assert device.has_lux
        assert device.temp == 12
        assert device.temp_unit == '°C'
        assert device.humidity == 100
        assert device.humidity_unit == '%'
        assert device.lux == 100
        assert device.lux_unit == 'lux'

    @requests_mock.mock()
    def tests_lm_float_units(self, m):
        """Tests that sensor/LM devices properties work as expected."""
        # Set up URL's
        m.post(CONST.LOGIN_URL, text=LOGIN.post_response_ok())
        m.get(CONST.OAUTH_TOKEN_URL, text=OAUTH_CLAIMS.get_response_ok())
        m.post(CONST.LOGOUT_URL, text=LOGOUT.post_response_ok())
        m.get(CONST.PANEL_URL, text=PANEL.get_response_ok(mode=CONST.MODE_STANDBY))
        m.get(
            CONST.DEVICES_URL,
            text=LM.device(
                devid=LM.DEVICE_ID,
                status='72.23 °F',
                temp='72.23 °F',
                lux='14.11 lx',
                humidity='34.38 %',
                low_battery=False,
                no_response=False,
            ),
        )

        # Logout to reset everything
        self.abode.logout()

        # Get our power switch
        device = self.abode.get_device(LM.DEVICE_ID)

        # Test our device
        assert device is not None
        assert device.status == '72.23 °F'
        assert not device.battery_low
        assert not device.no_response
        assert device.has_temp
        assert device.has_humidity
        assert device.has_lux
        assert device.temp == 72.23
        assert device.temp_unit == '°F'
        assert device.humidity == 34.38
        assert device.humidity_unit == '%'
        assert device.lux == 14.11
        assert device.lux_unit == 'lux'

    @requests_mock.mock()
    def tests_lm_temp_only(self, m):
        """Tests that sensor/LM devices properties work as expected."""
        # Set up URL's
        m.post(CONST.LOGIN_URL, text=LOGIN.post_response_ok())
        m.get(CONST.OAUTH_TOKEN_URL, text=OAUTH_CLAIMS.get_response_ok())
        m.post(CONST.LOGOUT_URL, text=LOGOUT.post_response_ok())
        m.get(CONST.PANEL_URL, text=PANEL.get_response_ok(mode=CONST.MODE_STANDBY))
        m.get(
            CONST.DEVICES_URL,
            text=LM.device(
                devid=LM.DEVICE_ID, status='72 °F', temp='72 °F', lux='', humidity=''
            ),
        )

        # Logout to reset everything
        self.abode.logout()

        # Get our power switch
        device = self.abode.get_device(LM.DEVICE_ID)

        # Test our device
        assert device is not None
        assert device.status == '72 °F'
        assert device.has_temp
        assert not device.has_humidity
        assert not device.has_lux
        assert device.temp == 72
        assert device.temp_unit == '°F'
        assert device.humidity is None
        assert device.humidity_unit is None
        assert device.lux is None
        assert device.lux_unit is None



"""Init file for devices directory."""
import json
import logging

from ..exceptions import AbodeException

from ..helpers import constants as CONST
from ..helpers import errors as ERROR


_LOGGER = logging.getLogger(__name__)


class AbodeDevice:
    """Class to represent each Abode device."""

    def __init__(self, json_obj, abode):
        """Set up Abode device."""
        self._json_state = json_obj
        self._device_id = json_obj.get('id')
        self._device_uuid = json_obj.get('uuid')
        self._name = json_obj.get('name')
        self._type = json_obj.get('type')
        self._type_tag = json_obj.get('type_tag')
        self._generic_type = json_obj.get('generic_type')
        self._abode = abode

        self._update_name()

    def set_status(self, status):
        """Set device status."""
        if self._json_state['control_url']:
            url = CONST.BASE_URL + self._json_state['control_url']

            status_data = {'status': str(status)}

            response = self._abode.send_request(method="put", url=url, data=status_data)
            response_object = json.loads(response.text)

            _LOGGER.debug("Set Status Response: %s", response.text)

            if response_object['id'] != self.device_id:
                raise AbodeException((ERROR.SET_STATUS_DEV_ID))

            if response_object['status'] != str(status):
                raise AbodeException((ERROR.SET_STATUS_STATE))

            # Note: Status result is of int type, not of new status of device.
            # Seriously, why would you do that?
            # So, can't set status here must be done at device level.

            _LOGGER.info("Set device %s status to: %s", self.device_id, status)

            return True

        return False

    def set_level(self, level):
        """Set device level."""
        if self._json_state['control_url']:
            url = CONST.BASE_URL + self._json_state['control_url']

            level_data = {'level': str(level)}

            response = self._abode.send_request("put", url, data=level_data)
            response_object = json.loads(response.text)

            _LOGGER.debug("Set Level Response: %s", response.text)

            if response_object['id'] != self.device_id:
                raise AbodeException((ERROR.SET_STATUS_DEV_ID))

            if response_object['level'] != str(level):
                raise AbodeException((ERROR.SET_STATUS_STATE))

            self.update(response_object)

            _LOGGER.info("Set device %s level to: %s", self.device_id, level)

            return True

        return False

    def get_value(self, name):
        """Get a value from the json object.

        This is the common data and is the best place to get state
        from if it has the data you require.
        This data is updated by the subscription service.
        """
        return self._json_state.get(name.lower(), {})

    def refresh(self, url=CONST.DEVICE_URL):
        """Refresh the devices json object data.

        Only needed if you're not using the notification service.
        """
        url = url.replace('$DEVID$', self.device_id)

        response = self._abode.send_request(method="get", url=url)
        response_object = json.loads(response.text)

        _LOGGER.debug("Device Refresh Response: %s", response.text)

        if response_object and not isinstance(response_object, (tuple, list)):
            response_object = [response_object]

        for device in response_object:
            self.update(device)

        return response_object

    def update(self, json_state):
        """Update the json data from a dictionary.

        Only updates if it already exists in the device.
        """
        self._json_state.update(
            {k: json_state[k] for k in json_state if self._json_state.get(k)}
        )
        self._update_name()

    def _update_name(self):
        """Set the device name from _json_state, with a sensible default."""
        self._name = self._json_state.get('name')
        if not self._name:
            self._name = self.type + ' ' + self.device_id

    @property
    def status(self):
        """Shortcut to get the generic status of a device."""
        return self.get_value('status')

    @property
    def battery_low(self):
        """Is battery level low."""
        return int(self.get_value('faults').get('low_battery', '0')) == 1

    @property
    def no_response(self):
        """Is the device responding."""
        return int(self.get_value('faults').get('no_response', '0')) == 1

    @property
    def out_of_order(self):
        """Is the device out of order."""
        return int(self.get_value('faults').get('out_of_order', '0')) == 1

    @property
    def tampered(self):
        """Has the device been tampered with."""
        # 'tempered' - Typo in API?
        return int(self.get_value('faults').get('tempered', '0')) == 1

    @property
    def name(self):
        """Get the name of this device."""
        return self._name

    @property
    def generic_type(self):
        """Get the generic type of this device."""
        return self._generic_type

    @property
    def type(self):
        """Get the type of this device."""
        return self._type

    @property
    def type_tag(self):
        """Get the type tag of this device."""
        return self._type_tag

    @property
    def device_id(self):
        """Get the device id."""
        return self._device_id

    @property
    def device_uuid(self):
        """Get the device uuid."""
        return self._device_uuid

    @property
    def desc(self):
        """Get a short description of the device."""
        # Garage Entry Door (ZW:00000003) - Door Lock - Closed
        return '{0} (ID: {1}, UUID: {2}) - {3} - {4}'.format(
            self.name, self.device_id, self.device_uuid, self.type, self.status
        )



"""Small SocketIO client via Websockets."""
import collections
import json
import logging
import threading

from datetime import datetime
from random import random

from lomond import WebSocket
from lomond import events
from lomond.persist import persist
from lomond.errors import WebSocketError

from .exceptions import SocketIOException
from .helpers import errors as ERRORS

STARTED = "started"
STOPPED = "stopped"
CONNECTED = "connected"
DISCONNECTED = "disconnected"
PING = "ping"
PONG = "pong"
POLL = "poll"
EVENT = "event"
ERROR = "error"

PACKET_OPEN = "0"
PACKET_CLOSE = "1"
PACKET_PING = "2"
PACKET_PONG = "3"
PACKET_MESSAGE = "4"

MESSAGE_CONNECT = "0"
MESSAGE_DISCONNECT = "1"
MESSAGE_EVENT = "2"
MESSAGE_ERROR = "4"

PING_INTERVAL = "pingInterval"
PING_TIMEOUT = "pingTimeout"

COOKIE_HEADER = str.encode("Cookie")
ORIGIN_HEADER = str.encode("Origin")

URL_PARAMS = "?EIO=3&transport=websocket"

_LOGGER = logging.getLogger(__name__)


class SocketIO:
    """Class for using websockets to talk to a SocketIO server."""

    def __init__(self, url, cookie=None, origin=None):
        """Init SocketIO class."""
        self._url = url + URL_PARAMS

        if origin:
            self._origin = origin.encode()
        else:
            self._origin = None

        if cookie:
            self._cookie = cookie.encode()
        else:
            self._cookie = None

        self._thread = None
        self._websocket = None
        self._exit_event = None
        self._running = False

        self._websocket_connected = False
        self._engineio_connected = False
        self._socketio_connected = False

        self._ping_interval_ms = 25000
        self._ping_timeout_ms = 60000

        self._last_ping_time = datetime.min
        self._last_packet_time = datetime.min

        self._callbacks = collections.defaultdict(list)

    def set_origin(self, origin=None):
        """Set the Origin header."""
        if origin:
            self._origin = origin.encode()
        else:
            self._origin = None

    def set_cookie(self, cookie=None):
        """Set the Cookie header."""
        if cookie:
            self._cookie = cookie.encode()
        else:
            self._cookie = None

    # pylint: disable=C0103
    def on(self, event_name, callback):
        """Register callback for a SocketIO event."""
        if not event_name:
            return False

        _LOGGER.debug("Adding callback for event name: %s", event_name)

        self._callbacks[event_name].append((callback))

        return True

    def start(self):
        """Start a thread to handle SocketIO notifications."""
        if not self._thread:
            _LOGGER.info("Starting SocketIO thread...")

            self._thread = threading.Thread(
                target=self._run_socketio_thread, name='SocketIOThread'
            )
            self._thread.deamon = True
            self._thread.start()

    def stop(self):
        """Tell the SocketIO thread to terminate."""
        if self._thread:
            _LOGGER.info("Stopping SocketIO thread...")

            # pylint: disable=W0212
            self._running = False

            if self._exit_event:
                self._exit_event.set()

            self._thread.join()

    def _run_socketio_thread(self):  # noqa: C901
        self._running = True

        # Back off for Error restarting
        min_wait = 5
        max_wait = 30

        retries = 0

        random_wait = max_wait - min_wait

        while self._running is True:
            _LOGGER.info("Attempting to connect to SocketIO server...")

            try:
                retries += 1

                self._handle_event(STARTED, None)

                self._websocket = WebSocket(self._url)
                self._exit_event = threading.Event()

                if self._cookie:
                    self._websocket.add_header(COOKIE_HEADER, self._cookie)

                if self._origin:
                    self._websocket.add_header(ORIGIN_HEADER, self._origin)

                for event in persist(
                    self._websocket, ping_rate=0, poll=5.0, exit_event=self._exit_event
                ):
                    if isinstance(event, events.Connected):
                        retries = 0
                        self._on_websocket_connected(event)
                    elif isinstance(event, events.Disconnected):
                        self._on_websocket_disconnected(event)
                    elif isinstance(event, events.Text):
                        self._on_websocket_text(event)
                    elif isinstance(event, events.Poll):
                        self._on_websocket_poll(event)
                    elif isinstance(event, events.BackOff):
                        self._on_websocket_backoff(event)

                    if self._running is False:
                        self._websocket.close()

            except SocketIOException as exc:
                _LOGGER.warning("SocketIO Error: %s", exc.details)

            except WebSocketError as exc:
                _LOGGER.warning("Websocket Error: %s", exc)

            if self._running:
                wait_for = min_wait + random() * min(random_wait, 2**retries)

                _LOGGER.info("Waiting %f seconds before reconnecting...", wait_for)

                if self._exit_event.wait(wait_for):
                    break

        self._handle_event(STOPPED, None)

    def _on_websocket_connected(self, _event):
        self._websocket_connected = True

        _LOGGER.info("Websocket Connected")

        self._handle_event(CONNECTED, None)

    def _on_websocket_disconnected(self, _event):
        self._websocket_connected = False
        self._engineio_connected = False
        self._socketio_connected = False

        _LOGGER.info("Websocket Disconnected")

        self._handle_event(DISCONNECTED, None)

    def _on_websocket_poll(self, _event):
        last_packet_delta = datetime.now() - self._last_packet_time
        last_packet_ms = int(last_packet_delta.total_seconds() * 1000)

        if self._engineio_connected and last_packet_ms > self._ping_timeout_ms:
            _LOGGER.warning("SocketIO Server Ping Timeout")
            self._websocket.close()
            return

        last_ping_delta = datetime.now() - self._last_ping_time
        last_ping_ms = int(last_ping_delta.total_seconds() * 1000)

        if self._engineio_connected and last_ping_ms >= self._ping_interval_ms:
            self._websocket.send_text(PACKET_PING)
            self._last_ping_time = datetime.now()
            _LOGGER.debug("Client Ping")
            self._handle_event(PING, None)

        self._handle_event(POLL, None)

    def _on_websocket_text(self, _event):
        self._last_packet_time = datetime.now()

        packet_type = _event.text[:1]
        packet_data = _event.text[1:]

        if packet_type == PACKET_OPEN:
            self._on_engineio_opened(packet_data)
        elif packet_type == PACKET_CLOSE:
            self._on_engineio_closed()
        elif packet_type == PACKET_PONG:
            self._on_engineio_pong()
        elif packet_type == PACKET_MESSAGE:
            self._on_engineio_message(packet_data)
        else:
            _LOGGER.debug("Ignoring EngineIO packet: %s", _event.text)

    # pylint: disable=R0201
    def _on_websocket_backoff(self, _event):
        return

    def _on_engineio_opened(self, _packet_data):
        json_data = json.loads(_packet_data)

        if json_data and json_data[PING_INTERVAL]:
            ping_interval_ms = json_data[PING_INTERVAL]
            _LOGGER.debug("Set ping interval to: %d", ping_interval_ms)

        if json_data and json_data[PING_TIMEOUT]:
            ping_timeout_ms = json_data[PING_TIMEOUT]
            _LOGGER.debug("Set ping timeout to: %d", ping_timeout_ms)

        self._engineio_connected = True

        _LOGGER.debug("EngineIO Connected")

    def _on_engineio_closed(self):
        self._engineio_connected = False

        _LOGGER.debug("EngineIO Disconnected")

        self._websocket.close()

    def _on_engineio_pong(self):
        _LOGGER.debug("Server Pong")
        self._handle_event(PONG, None)

    def _on_engineio_message(self, _packet_data):
        message_type = _packet_data[:1]
        message_data = _packet_data[1:]

        if message_type == MESSAGE_CONNECT:
            self._on_socketio_connected()
        elif message_type == MESSAGE_DISCONNECT:
            self._on_socketio_disconnected()
        elif message_type == MESSAGE_ERROR:
            self._on_socketio_error(message_data)
        elif message_type == MESSAGE_EVENT:
            self._on_socketio_event(message_data)
        else:
            _LOGGER.debug("Ignoring SocketIO message: %s", _packet_data)

    def _on_socketio_connected(self):
        self._socketio_connected = True

        _LOGGER.debug("SocketIO Connected")

    def _on_socketio_disconnected(self):
        self._socketio_connected = False

        _LOGGER.debug("SocketIO Disconnected")

        self._websocket.close()

    def _on_socketio_error(self, _message_data):
        self._handle_event(ERROR, _message_data)

        raise SocketIOException(ERRORS.SOCKETIO_ERROR, details=_message_data)

    def _on_socketio_event(self, _message_data):
        l_bracket = _message_data.find("[")
        r_bracket = _message_data.rfind("]")

        if l_bracket == -1 or r_bracket == -1:
            _LOGGER.warning("Unable to find event [data]: %s", _message_data)
            return

        json_str = _message_data[l_bracket : r_bracket + 1]
        json_data = json.loads(json_str)

        self._handle_event(EVENT, _message_data)
        self._handle_event(json_data[0], json_data[1:])

    def _handle_event(self, event_name, event_data):
        for callback in self._callbacks.get(event_name, ()):
            try:
                if event_data:
                    callback(event_data)
                else:
                    callback()
            # pylint: disable=W0703
            except Exception as exc:
                _LOGGER.exception(
                    "Captured exception during SocketIO event callback: %s", exc
                )



"""Mock Abode IP Camera Device."""
import jaraco.abode.helpers.constants as CONST

DEVICE_ID = 'ZB:00000305'
CONTROL_URL = 'api/v1/cams/' + DEVICE_ID + '/record'
CONTROL_URL_SNAPSHOT = 'api/v1/cams/' + DEVICE_ID + '/capture'


def device(
    devid=DEVICE_ID,
    status=CONST.STATUS_ONLINE,
    low_battery=False,
    no_response=False,
    privacy=1,
):
    """IP camera mock device."""
    return (
        '''
        {
          "id":"'''
        + devid
        + '''",
          "type_tag": "device_type.ipcam",
          "type": "IP Cam",
          "name": "Living Room Camera",
          "area": "1",
          "zone": "1",
          "sort_order": "",
          "is_window": "",
          "bypass": "0",
          "schar_24hr": "1",
          "sresp_24hr": "5",
          "sresp_mode_0": "0",
          "sresp_entry_0": "0",
          "sresp_exit_0": "0",
          "group_name": "Streaming Camera",
          "group_id": "397974",
          "default_group_id": "1",
          "sort_id": "10000",
          "sresp_mode_1": "0",
          "sresp_entry_1": "0",
          "sresp_exit_1": "0",
          "sresp_mode_2": "0",
          "sresp_entry_2": "0",
          "sresp_exit_2": "0",
          "sresp_mode_3": "0",
          "uuid": "123456789",
          "sresp_entry_3": "0",
          "sresp_exit_3": "0",
          "sresp_mode_4": "0",
          "sresp_entry_4": "0",
          "sresp_exit_4": "0",
          "version": "1.0.2.22G_6.8E_homekit_2.0.9_s2 ABODE oz",
          "origin": "abode",
          "has_subscription": null,
          "onboard": "1",
          "s2_grnt_keys": "",
          "s2_dsk": "",
          "s2_propty": "",
          "s2_keys_valid": "",
          "zwave_secure_protocol": "",
          "control_url":"'''
        + CONTROL_URL
        + '''",
          "deep_link": null,
          "status_color": "#5cb85c",
          "faults": {
              "low_battery":'''
        + str(int(low_battery))
        + ''',
              "tempered": 0,
              "supervision": 0,
              "out_of_order": 0,
              "no_response":'''
        + str(int(no_response))
        + ''',
              "jammed": 0,
              "zwave_fault": 0
          },
          "status":"'''
        + status
        + '''",
          "status_display": "Online",
          "statuses": [],
          "status_ex": "",
          "actions": [
              {
                  "label": "Capture Video",
                  "value": "a=1&z=1&req=vid;"
              },
              {
                  "label": "Turn off Live Video",
                  "value": "a=1&z=1&privacy=on;"
              },
              {
                  "label": "Turn on Live Video",
                  "value": "a=1&z=1&privacy=off;"
              }
          ],
          "status_icons": [],
          "icon": "assets/icons/streaming-camaera-new.svg",
          "control_url_snapshot":"'''
        + CONTROL_URL_SNAPSHOT
        + '''",
          "ptt_supported": true,
          "is_new_camera": 1,
          "stream_quality": 3,
          "camera_mac": "AB:CD:EF:GF:HI",
          "privacy":"'''
        + str(privacy)
        + '''",
          "enable_audio": "1",
          "alarm_video": "25",
          "pre_alarm_video": "5",
          "mic_volume": "75",
          "speaker_volume": "75",
          "mic_default_volume": 40,
          "speaker_default_volume": 46,
          "bandwidth": {
              "slider_labels": [
                  {
                      "name": "High",
                      "value": 3
                  },
                  {
                      "name": "Medium",
                      "value": 2
                  },
                  {
                      "name": "Low",
                      "value": 1
                  }
              ],
              "min": 1,
              "max": 3,
              "step": 1
          },
          "volume": {
              "min": 0,
              "max": 100,
              "step": 1
          },
          "video_flip": "0",
          "hframe": "1080P"
        }'''
    )


def get_capture_timeout():
    """Mock timeout response."""
    return '''
    {
        "code":600,
        "message":"Image Capture request has timed out.",
        "title":"",
        "detail":null
    }'''


FILE_PATH_ID = 'ZB00000305'
FILE_PATH = 'api/storage/' + FILE_PATH_ID + '/2020-01-26/173238/0.jpg'

LOCATION_HEADER = (
    'https://www.google.com/images/branding/googlelogo/'
    + '1x/googlelogo_color_272x92dp.png'
)


def timeline_event(devid=DEVICE_ID, event_code='5001', file_path=FILE_PATH):
    """Camera Timeline Event Mockup."""
    return (
        '''
    {
        "mac": "B0:C5:CZ:54:12:9A",
        "id": "1171272698",
        "xml": null,
        "date": "01/26/2020",
        "time": "05:32 PM",
        "event_utc": "1580088758",
        "event_cid": "",
        "event_code": "'''
        + event_code
        + '''",
        "device_id": "'''
        + devid
        + '''",
        "device_type_id": "69",
        "device_type": "IP Cam",
        "timeline_ha_device": null,
        "d_name": "Living Room Camera",
        "delete_by_user": null,
        "pin_code_user": " ",
        "file_del_at": "",
        "nest_has_motion": null,
        "nest_has_sound": null,
        "nest_has_person": null,
        "neaz": null,
        "hasFaults": "0",
        "file_path":"'''
        + file_path
        + '''",
        "deep_link": null,
        "file_name": "48755_b0c5ca37894b_2020-01-26_173238_0-M2+56431.jpg",
        "file_size": "197207",
        "file_count": "1",
        "file_is_del": "0",
        "event_type": "Image Capture",
        "severity": "6",
        "pos": "l",
        "color": "#40bbea",
        "is_alarm": "0",
        "triggered_by_str": null,
        "ha_type": null,
        "ha_device_name": null,
        "ha_location": null,
        "ha_mobile": null,
        "ha_cond": null,
        "h_location": null,
        "ha_trigger": null,
        "icon": "assets/email/motion-camera.png",
        "user_id": "95244",
        "user_name": "Shred",
        "mobile_name": "",
        "parent_tid": "",
        "app_type": "WebApp",
        "viewed_by_uid": null,
        "verified_by_tid": null,
        "la_applied_by": null,
        "la_event_type": null,
        "la_culprit_mobiles": null,
        "la_executed": null,
        "la_applied_at": null,
        "device_name": "Living Room Camera",
        "event_name": "Living Room Camera Image Capture",
        "event_by": "by Shred using WebApp",
        "file_delete_by": ""
    }'''
    )



"""Abode cloud push events."""
import collections
import logging

from .devices import AbodeDevice
from .exceptions import AbodeException
from .helpers import constants as CONST
from .helpers import errors as ERROR
from .helpers import timeline as TIMELINE
from . import socketio as sio

_LOGGER = logging.getLogger(__name__)


class AbodeEventController:
    """Class for subscribing to abode events."""

    def __init__(self, abode, url=CONST.SOCKETIO_URL):
        """Init event subscription class."""
        self._abode = abode
        self._thread = None
        self._running = False
        self._connected = False

        # Setup callback dicts
        self._connection_status_callbacks = collections.defaultdict(list)
        self._device_callbacks = collections.defaultdict(list)
        self._event_callbacks = collections.defaultdict(list)
        self._timeline_callbacks = collections.defaultdict(list)

        # Setup SocketIO
        self._socketio = sio.SocketIO(url=url, origin=CONST.BASE_URL)

        # Setup SocketIO Callbacks
        self._socketio.on(sio.STARTED, self._on_socket_started)
        self._socketio.on(sio.CONNECTED, self._on_socket_connected)
        self._socketio.on(sio.DISCONNECTED, self._on_socket_disconnected)
        self._socketio.on(CONST.DEVICE_UPDATE_EVENT, self._on_device_update)
        self._socketio.on(CONST.GATEWAY_MODE_EVENT, self._on_mode_change)
        self._socketio.on(CONST.TIMELINE_EVENT, self._on_timeline_update)
        self._socketio.on(CONST.AUTOMATION_EVENT, self._on_automation_update)

    def start(self):
        """Start a thread to handle Abode SocketIO notifications."""
        self._socketio.start()

    def stop(self):
        """Tell the subscription thread to terminate - will block."""
        self._socketio.stop()

    def add_connection_status_callback(self, unique_id, callback):
        """Register callback for Abode server connection status."""
        if not unique_id:
            return False

        _LOGGER.debug("Subscribing to Abode connection updates for: %s", unique_id)

        self._connection_status_callbacks[unique_id].append((callback))

        return True

    def remove_connection_status_callback(self, unique_id):
        """Unregister connection status callbacks."""
        if not unique_id:
            return False

        _LOGGER.debug("Unsubscribing from Abode connection updates for : %s", unique_id)

        self._connection_status_callbacks[unique_id].clear()

        return True

    def add_device_callback(self, devices, callback):
        """Register a device callback."""
        if not devices:
            return False

        if not isinstance(devices, (tuple, list)):
            devices = [devices]

        for device in devices:
            # Device may be a device_id
            device_id = device

            # If they gave us an actual device, get that devices ID
            if isinstance(device, AbodeDevice):
                device_id = device.device_id

            # Validate the device is valid
            if not self._abode.get_device(device_id):
                raise AbodeException((ERROR.EVENT_DEVICE_INVALID))

            _LOGGER.debug("Subscribing to updates for device_id: %s", device_id)

            self._device_callbacks[device_id].append((callback))

        return True

    def remove_all_device_callbacks(self, devices):
        """Unregister all callbacks for a device."""
        if not devices:
            return False

        if not isinstance(devices, (tuple, list)):
            devices = [devices]

        for device in devices:
            device_id = device

            if isinstance(device, AbodeDevice):
                device_id = device.device_id

            if not self._abode.get_device(device_id):
                raise AbodeException((ERROR.EVENT_DEVICE_INVALID))

            if device_id not in self._device_callbacks:
                return False

            _LOGGER.debug("Unsubscribing from all updates for device_id: %s", device_id)

            self._device_callbacks[device_id].clear()

        return True

    def add_event_callback(self, event_groups, callback):
        """Register callback for a group of timeline events."""
        if not event_groups:
            return False

        if not isinstance(event_groups, (tuple, list)):
            event_groups = [event_groups]

        for event_group in event_groups:
            if event_group not in TIMELINE.ALL_EVENT_GROUPS:
                raise AbodeException(
                    ERROR.EVENT_GROUP_INVALID, TIMELINE.ALL_EVENT_GROUPS
                )

            _LOGGER.debug("Subscribing to event group: %s", event_group)

            self._event_callbacks[event_group].append((callback))

        return True

    def add_timeline_callback(self, timeline_events, callback):
        """Register a callback for a specific timeline event."""
        if not timeline_events:
            return False

        if not isinstance(timeline_events, (tuple, list)):
            timeline_events = [timeline_events]

        for timeline_event in timeline_events:
            if not isinstance(timeline_event, dict):
                raise AbodeException((ERROR.EVENT_CODE_MISSING))

            event_code = timeline_event.get('event_code')

            if not event_code:
                raise AbodeException((ERROR.EVENT_CODE_MISSING))

            _LOGGER.debug("Subscribing to timeline event: %s", timeline_event)

            self._timeline_callbacks[event_code].append((callback))

        return True

    @property
    def connected(self):
        """Get the Abode connection status."""
        return self._connected

    @property
    def socketio(self):
        """Get the SocketIO instance."""
        return self._socketio

    def _on_socket_started(self):
        """Socket IO startup callback."""
        # pylint: disable=W0212
        cookies = self._abode._get_session().cookies.get_dict()
        cookie_string = "; ".join([str(x) + "=" + str(y) for x, y in cookies.items()])

        self._socketio.set_cookie(cookie_string)

    def _on_socket_connected(self):
        """Socket IO connected callback."""
        self._connected = True

        try:
            self._abode.refresh()
        # pylint: disable=W0703
        except Exception as exc:
            _LOGGER.warning("Captured exception during Abode refresh: %s", exc)
        finally:
            # Callbacks should still execute even if refresh fails (Abode
            # server issues) so that the entity availability in Home Assistant
            # is updated since we are in fact connected to the web socket.
            for callbacks in self._connection_status_callbacks.items():
                for callback in callbacks[1]:
                    _execute_callback(callback)

    def _on_socket_disconnected(self):
        """Socket IO disconnected callback."""
        self._connected = False

        for callbacks in self._connection_status_callbacks.items():
            # Check if list is not empty.
            # Applicable when remove_all_device_callbacks
            # is called before _on_socket_disconnected.
            if callbacks[1]:
                for callback in callbacks[1]:
                    _execute_callback(callback)

    def _on_device_update(self, devid):
        """Device callback from Abode SocketIO server."""
        if isinstance(devid, (tuple, list)):
            devid = devid[0]

        if devid is None:
            _LOGGER.warning("Device update with no device id.")
            return

        _LOGGER.debug("Device update event for device ID: %s", devid)

        device = self._abode.get_device(devid, True)

        if not device:
            _LOGGER.debug("Got device update for unknown device: %s", devid)
            return

        for callback in self._device_callbacks.get(device.device_id, ()):
            _execute_callback(callback, device)

    def _on_mode_change(self, mode):
        """Mode change broadcast from Abode SocketIO server."""
        if isinstance(mode, (tuple, list)):
            mode = mode[0]

        if mode is None:
            _LOGGER.warning("Mode change event with no mode.")
            return

        if not mode or mode.lower() not in CONST.ALL_MODES:
            _LOGGER.warning("Mode change event with unknown mode: %s", mode)
            return

        _LOGGER.debug("Alarm mode change event to: %s", mode)

        # We're just going to convert it to an Alarm device
        alarm_device = self._abode.get_alarm(refresh=True)

        # At the time of development, refreshing after mode change notification
        # didn't seem to get the latest update immediately. As such, we will
        # force the mode status now to match the notification.
        # pylint: disable=W0212
        alarm_device._json_state['mode']['area_1'] = mode

        for callback in self._device_callbacks.get(alarm_device.device_id, ()):
            _execute_callback(callback, alarm_device)

    def _on_timeline_update(self, event):
        """Timeline update broadcast from Abode SocketIO server."""
        if isinstance(event, (tuple, list)):
            event = event[0]

        event_type = event.get('event_type')
        event_code = event.get('event_code')

        if not event_type or not event_code:
            _LOGGER.warning("Invalid timeline update event: %s", event)
            return

        _LOGGER.debug(
            "Timeline event received: %s - %s (%s)",
            event.get('event_name'),
            event_type,
            event_code,
        )

        # Compress our callbacks into those that match this event_code
        # or ones registered to get callbacks for all events
        codes = (event_code, TIMELINE.ALL['event_code'])
        all_callbacks = [self._timeline_callbacks[code] for code in codes]

        for callbacks in all_callbacks:
            for callback in callbacks:
                _execute_callback(callback, event)

        # Attempt to map the event code to a group and callback
        event_group = TIMELINE.map_event_code(event_code)

        if event_group:
            for callback in self._event_callbacks.get(event_group, ()):
                _execute_callback(callback, event)

    def _on_automation_update(self, event):
        """Automation update broadcast from Abode SocketIO server."""
        event_group = TIMELINE.AUTOMATION_EDIT_GROUP

        if isinstance(event, (tuple, list)):
            event = event[0]

        for callback in self._event_callbacks.get(event_group, ()):
            _execute_callback(callback, event)


def _execute_callback(callback, *args, **kwargs):
    # Callback with some data, capturing any exceptions to prevent chaos
    try:
        callback(*args, **kwargs)
    # pylint: disable=W0703
    except Exception as exc:
        _LOGGER.warning("Captured exception during callback: %s", exc)



"""Mock Abode Power Switch Sensor Device."""
import jaraco.abode.helpers.constants as CONST

DEVICE_ID = 'ZW:00000105'
CONTROL_URL = 'api/v1/control/light/' + DEVICE_ID


def device(
    devid=DEVICE_ID,
    status=CONST.STATUS_OFF,
    level=0,
    low_battery=False,
    no_response=False,
):
    """Dimmer  mock device."""
    return (
        '''
      {
         "id":"'''
        + devid
        + '''",
         "type_tag":"device_type.dimmer",
         "type":"Dimmer",
         "name":"Kitchen Lights",
         "area":"1",
         "zone":"12",
         "sort_order":"",
         "is_window":"",
         "bypass":"0",
         "schar_24hr":"0",
         "sresp_24hr":"0",
         "sresp_mode_0":"0",
         "sresp_entry_0":"0",
         "sresp_exit_0":"0",
         "group_name":"Ungrouped",
         "group_id":"1",
         "default_group_id":"1",
         "sort_id":"10000",
         "sresp_mode_1":"0",
         "sresp_entry_1":"0",
         "sresp_exit_1":"0",
         "sresp_mode_2":"0",
         "sresp_entry_2":"0",
         "sresp_exit_2":"0",
         "sresp_mode_3":"0",
         "uuid":"fcc65fb7d52cdf7b080a2005539e30a4",
         "sresp_entry_3":"0",
         "sresp_exit_3":"0",
         "sresp_mode_4":"0",
         "sresp_entry_4":"0",
         "sresp_exit_4":"0",
         "version":"",
         "origin":"abode",
         "has_subscription":null,
         "onboard":"0",
         "s2_grnt_keys":"",
         "s2_dsk":"",
         "s2_propty":"",
         "s2_keys_valid":"",
         "zwave_secure_protocol":"",
         "control_url":"'''
        + CONTROL_URL
        + '''",
         "deep_link":null,
         "status_color":"#5cb85c",
         "faults":{
            "low_battery":'''
        + str(int(low_battery))
        + ''',
            "tempered":0,
            "supervision":0,
            "out_of_order":0,
            "no_response":'''
        + str(int(no_response))
        + ''',
            "jammed":0,
            "zwave_fault":0
         },
         "status":"'''
        + status
        + '''",
         "status_display":"OFF",
         "statuses":{
            "saturation":"N/A",
            "hue":"N/A",
            "level":"'''
        + str(int(level))
        + '''",
            "switch":"0",
            "color_temp":"N/A",
            "color_mode":"N/A"
         },
         "status_ex":"",
         "actions":[
            {
               "label":"Switch off",
               "value":"a=1&z=12&sw=off;"
            },
            {
               "label":"Switch on",
               "value":"a=1&z=12&sw=on;"
            },
            {
               "label":"Toggle",
               "value":"a=1&z=12&sw=toggle;"
            },
            {
               "label":"0%",
               "value":"a=1&z=12&sw=0;"
            },
            {
               "label":"10%",
               "value":"a=1&z=12&sw=10;"
            },
            {
               "label":"20%",
               "value":"a=1&z=12&sw=20;"
            },
            {
               "label":"30%",
               "value":"a=1&z=12&sw=30;"
            },
            {
               "label":"40%",
               "value":"a=1&z=12&sw=40;"
            },
            {
               "label":"50%",
               "value":"a=1&z=12&sw=50;"
            },
            {
               "label":"60%",
               "value":"a=1&z=12&sw=60;"
            },
            {
               "label":"70%",
               "value":"a=1&z=12&sw=70;"
            },
            {
               "label":"80%",
               "value":"a=1&z=12&sw=80;"
            },
            {
               "label":"90%",
               "value":"a=1&z=12&sw=90;"
            },
            {
               "label":"100%",
               "value":"a=1&z=12&sw=99;"
            }
         ],
         "status_icons":[

         ],
         "statusEx":"0",
         "icon":"assets/icons/bulb-1.svg"
      }'''
    )
